"""
Provide an abstract product locator for GOES-R Series imagery dataset.

Classes:
    GOESProductLocator: Abstract a product locator for GOES-R Series
    imagery dataset.
"""

from datetime import datetime, timedelta
from typing import ClassVar

from ..dataset import ProductLocatorGG
from .constants import (
    GOESR_FILE_DATE_FORMAT,
    GOESR_FILE_DATE_PATTERN,
    GOESR_FILE_SUFFIX,
    GOESR_PATH_DATE_FORMAT,
)


class GOESProductLocator(ProductLocatorGG):
    """
    Abstract a product locator for GOES-R Series imagery dataset.

    This abstract class implements the `ProductLocator` interface for
    specific GOES-R Series imagery dataset product locators by
    inheriting from the `ProductLocatorGG` abstract class. The class
    implements the common specifications and naming conventions for
    products in the GOES-R dataset, and serves as a base class for more
    specialised product locator classes. It also defines the common
    attributes and abstract methods for all GOES-R Series imagery
    dataset product locators.

    Instances of this class are responsible for generating a list of
    folder paths based on the dataset's directory structure and naming
    conventions, product details, and the specified date range. The
    generated paths cover each hour interval within the required period
    according on the dataset's temporal granularity; paths to the
    folders containing the initial and final dates are included in the
    list.

    `GOESProductLocator` objects are responsible for verifying if a
    given filename matches the product filename pattern based on the
    dataset's naming conventions and product specifications using the
    `match()` method, and for extracting the corresponding `datetime`
    information from the product's filename by dint of the method
    `get_datetime()`.

    Also, the class provides the implementation for the abstract helper
    methods defined in the `ProductLocatorGG` abstract class. The
    `GOESProductLocator` class is the workhorse for all GOES-R Series
    imagery dataset product locators.

    Notes
    -----
    Subclasses are responsible for initialising the attributes with the
    appropriate values for the dataset and product details.

    Methods
    -------
    get_base_url(datasource: str)
        Get the base URL for the GOES-R Series imagery dataset's
        products.
    get_date_format()
        Return the date format specification for the GOES-R Series
        imagery dataset product's filename.
    get_paths(datetime_ini: datetime, datetime_fin: datetime) -> list[str]:
        Generate a list of paths containing the product files for the
        specified date range.
    get_prefix()
        Generate the prefix for the GOES-R Series imagery dataset
        product's filename.
    get_product_tag()
        Generate the product prefix for the GOES-R Series imagery
        dataset product's filename.
    get_mode_tag()
        Generate the scan mode and channel number identifier part of the
        GOES-R Series imagery dataset product's filename.
    get_suffix()
        Generate the suffix for the GOES-R Series imagery dataset
        product's filename.
    get_timestamp_pattern()
        Return the timestamp pattern for the GOES-R Series imagery
        dataset product's filename.
    next_time(current_time: datetime)
        Get the next time interval. GOES-R Series dataset organises the
        data by hour.
    normalize_times(datetime_ini: datetime, datetime_fin: datetime)
        Normalise the initial and final datetimes.
    truncate_to_hour(time: datetime)
        Truncate the `datetime` to the current hour.

    Caution
    -------
    Members of this class not defined by the `ProductLocator` interface
    are helper methods and can be considered as implementation details,
    even though they are defined as part of the public API. In future
    releases, these methods may be moved to a private scope, suffer name
    changes, or be removed altogether.
    """  # noqa: E501

    # Supported datasources of the GOES-R Series imagery dataset
    # products:
    SUPPORTED_DATASOURCES: ClassVar[set[str]] = {"AWS"}

    # Satellites in the GOES-R Series are identified by the following
    # IDs:
    AVAILABLE_ORIGINS: dict[str, str] = {
        f"G{idn:02d}": f"goes{idn:02d}" for idn in range(16, 20)
    }

    # Supported instruments from the GOES-R series:
    AVAILABLE_INSTRUMENTS: dict[str, str] = {
        "ABI": "Advanced Baseline Imager",
        "GLM": "Geostationary Lightning Mapper",
    }

    # Supported product levels from the GOES-R series:
    # - Level 1b (calibrated and geographically corrected, radiance
    #             units)
    # - Level 2  (calibrated and geographically corrected,
    #             reflectance/brightness [Kelvin] units)
    AVAILABLE_LEVELS: set[str] = {"L1b", "L2"}

    FULL_DISK = "F"
    CONUS = "C"
    MESO = "M"
    MESO_1 = "M1"
    MESO_2 = "M2"

    # WARNING: This is being override by GOESProductLocatorABI in this
    #          release due to an incomplete refactoring. This will be
    #          fixed in future releases. Nevertheless, the methods that
    #          use this attribute was fixed to use the correct values.
    AVAILABLE_SCENES: dict[str, str] = {
        FULL_DISK: FULL_DISK,
        CONUS: CONUS,
        MESO_1: MESO,
        MESO_2: MESO,
    }

    def __init__(
        self,
        name: str,
        level: str,
        scene: str,
        instrument: str,
        modes: list[str],
        channels: list[str],
        origin: str,
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset product locator.

        Constructs a GOES-R Series imagery dataset product locator
        object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        level : str
            The level of the GOES-R Series imagery dataset product, e.g.
            "L1b" or "L2".
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        instrument : str
            The instrument of the GOES-R Series imagery dataset product,
            e.g. "ABI" or "GLM".
        modes : list[str]
            The list of modes of the GOES-R Series imagery dataset
            product, e.g. "M3" or "M6".
        channels : list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Notes
        -----
        ValueError is raised if the provided origin, level or instrument
        is invalid. Or if an unexpected or unsupported setting is
        required for an instrument that does not support it.
        """
        # TODO: Too many positional arguments. Solve it by using
        #       the Builder or Factory methods, or patterns like
        #       Essence or Fluent.
        self._validate_origin(origin, self.AVAILABLE_ORIGINS)
        self._validate_instrument(instrument, self.AVAILABLE_INSTRUMENTS)
        self._validate_level(level, self.AVAILABLE_LEVELS)

        self.name: str = name
        self.level: str = level
        self.scene: str = scene
        self.instrument: str = instrument
        self.modes: list[str] = modes
        self.channels: list[str] = channels
        self.origin: str = origin

    def get_base_url(self, datasource: str) -> tuple[str, ...]:
        """
        Get the base URL for the GOES-R Series dataset's products.

        This method returns the base URL for the GOES-R Series imagery
        dataset's products. The base URL is used to construct the full
        URL to the dataset's product files.

        Parameters
        ----------
        datasource : str
            The datasource identifier. This parameter is used to
            determine the base URL for the dataset's products.

        Returns
        -------
        tuple[str, ...]
            A tuple whose first element is the base URL for the GOES-R
            Series imagery dataset's products based on the requested
            datasource identifier. The second element is an empty
            string.

        Raises
        ------
        ValueError
            If the provided datasource is not supported or unavailable.
        """
        if datasource not in self.SUPPORTED_DATASOURCES:
            supported_datasources: list[str] = sorted(
                self.SUPPORTED_DATASOURCES
            )
            raise ValueError(
                f"Unsupported datasource: '{datasource}'. "
                f"Supported datasources: {supported_datasources}"
            )

        scene: str = self.scene
        product: str = f"{self.instrument}-{self.level}-{self.name}{scene}"
        satellite: str = self.AVAILABLE_ORIGINS[self.origin]

        available_datasources: dict[str, str] = {
            "AWS": f"s3://noaa-{satellite}/{product}/"
        }

        return (available_datasources[datasource], "")

    def get_date_format(self) -> str:
        """
        Return the date format specification for the product's filename.

        Generates and returns the date format specification for the
        product's filename based on the GOES-R Series imagery dataset
        product filename's date and time format conventions. The date
        format specification string is used to parse the product's
        filename and extract the `datetime` information.

        Returns
        -------
        str
            The date format specification for the GOES-R Series imagery
            dataset product's filename.
        """
        return GOESR_FILE_DATE_FORMAT

    def get_paths(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> list[str]:
        """
        Generate a list of dataset directory paths.

        This method generates a list of directory paths within the
        dataset based on the folder structure and naming conventions,
        temporal granularity, and the specified date range. Paths to the
        folders containing the initial and final dates are included in
        the list.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime for the desired data.
        datetime_fin : datetime
            The final datetime for the desired data.

        Returns
        -------
        list[str]
            A list of strings representing the paths to the dataset
            directories containing the product files for the specified
            date range.
        """
        current_time: datetime
        end_time: datetime
        current_time, end_time = self.normalize_times(
            datetime_ini, datetime_fin
        )

        folder_paths: list[str] = []

        while current_time <= end_time:
            folder_path: str = current_time.strftime(GOESR_PATH_DATE_FORMAT)
            folder_paths.append(f"{folder_path}/")

            current_time = self.next_time(current_time)

        return folder_paths

    def get_prefix(self) -> str:
        """
        Return the prefix for the product's filename.

        Generates and returns the prefix for the GOES-R Series imagery
        dataset product's filename based on product-specific information
        like dataset and product's name, instrument and origin's
        identifier, etc.

        Returns
        -------
        str
            The prefix for the GOES-R Series imagery dataset product's
            filename.
        """
        product_prefix: str = self.get_product_tag()
        scan_band: str = self.get_mode_tag()

        if scan_band:
            scan_band = f"-{scan_band}"
        origin: str = f"_{self.origin}"

        return f"OR_{product_prefix}{scan_band}{origin}"

    def get_product_tag(self) -> str:
        """
        Return a product identifier for the product's filename.

        Generates and returns the product identifier to be used as part
        of the prefix for the GOES-R Series imagery dataset product's
        filename based on information like product's name, instrument
        and origin's identifier, etc.

        Returns
        -------
        str
            The product identifier for the GOES-R Series imagery dataset
            product's filename.
        """
        product_id: str = f"{self.level}-{self.name}{self.scene}"

        return f"{self.instrument}-{product_id}"

    def get_mode_tag(self) -> str:
        """
        Return the scan mode and band names for the product's filename.

        Generates and returns a regex pattern for the requested scan
        modes and channels that appear as tags in the product's filename
        based on product-specific scan modes and band channel
        identifiers.

        Returns
        -------
        str
            The scan mode and band names for the GOES-R Series imagery
            dataset product's filename.
        """
        sorted_modes: list[str] = sorted(self.modes)
        modes: str = f"(?:{'|'.join(sorted_modes)})" if self.modes else ""

        sorted_channels: list[str] = sorted(self.channels)
        channels: str = (
            f"(?:{'|'.join(sorted_channels)})" if self.channels else ""
        )

        return f"{modes}{channels}"

    def get_suffix(self) -> str:
        """
        Return the suffix for the product's filename.

        Generates and returns the suffix for the product's filename
        based on product-specific information like product's version
        origin's identifier, and file suffix (extension).

        Returns
        -------
        str
            The suffix for the GOES-R Series imagery dataset product's
            filename.
        """
        return GOESR_FILE_SUFFIX

    def get_timestamp_pattern(self) -> str:
        """
        Return the timestamp regex pattern for the product's filename.

        Generates and returns the timestamp regular expression
        pattern for the product's filename based on the dataset
        product filename's date and time format conventions. The
        timestamp regex pattern is used to extract the substring
        containing the timestamp from the product's filename
        before extracting the `datetime` information.

        Returns
        -------
        str
            The timestamp regex pattern for the GOES-R Series imagery
            dataset product's filename.
        """
        start_date: str = f"_s({GOESR_FILE_DATE_PATTERN})"
        end_date: str = f"_e{GOESR_FILE_DATE_PATTERN}"
        creation_date: str = f"_c{GOESR_FILE_DATE_PATTERN}"

        return f"{start_date}{end_date}{creation_date}"

    @staticmethod
    def next_time(current_time: datetime) -> datetime:
        """
        Get the next time interval.

        Get the next time interval based on the current time interval.

        Parameters
        ----------
        current_time : datetime
            The current time interval.

        Returns
        -------
        datetime
            The next time interval.
        """
        return current_time + timedelta(hours=1)

    def normalize_times(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> tuple[datetime, datetime]:
        """
        Normalise the initial and final datetimes.

        Normalise the initial and final datetimes to the nearest
        time interval based on the dataset. The initial datetime is
        normalised to the start of the time interval and the final
        datetime is normalised to the end of the time interval.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime.
        datetime_fin : datetime
            The final datetime.

        Returns
        -------
        tuple[datetime, datetime]
            A tuple containing the normalised initial and final
            datetimes.
        """
        start_time: datetime = self.truncate_to_hour(datetime_ini)
        end_time: datetime = self.truncate_to_hour(datetime_fin)

        return start_time, end_time

    @staticmethod
    def truncate_to_hour(time: datetime) -> datetime:
        """
        Truncate the `datetime` to the current hour.

        The `datetime` is truncated to the beginning of the current
        hour.

        Parameters
        ----------
        time : datetime
            The `datetime` to be truncated to the current hour.

        Returns
        -------
        datetime
            The `datetime` truncated to the current hour.
        """
        return time.replace(minute=0, second=0, microsecond=0)



"""
Provide methods for handling IR brightness temperature color enhancement.

It includes functions to load, parse, process, and save McIDAS and GMT
enhancement color tables, as well as reverse and manipulate color maps.
"""

from .catalogue import cmap
from .colormap import EnhancementColormap
from .palette import EnhacementPalette
from .preview import show_colormap
from .scale import EnhancementScale
from .stretching import EnhacementStretching
from .table import EnhacementTable

__all__ = [
    "EnhancementColormap",
    "EnhacementPalette",
    "EnhancementScale",
    "EnhacementStretching",
    "EnhacementTable",
    "show_colormap",
    "cmap",
]



"""
Provide a locator for the GridSat-GOES/CONUS imagery dataset products.

Attributes
----------
MONTHS_IN_YEAR : int
    The number of months in a year, used for date calculations.

Classes
-------
GridSatProductLocatorGC
    Represent the GridSat-GOES/CONUS imagery dataset product locator.
"""

from datetime import datetime

from .constants import (
    GOES_FILE_DATE_FORMAT,
    GOES_FILE_DATE_PATTERN,
    GOES_FILE_PREFIX,
    GOES_PATH_DATE_FORMAT,
    GOES_PRODUCT_LATEST_VERSION,
)
from .locator import GridSatProductLocator

# Define a constant for the number of months in a year
MONTHS_IN_YEAR = 12


class GridSatProductLocatorGC(GridSatProductLocator):
    """
    Represent the GridSat-GOES/CONUS imagery dataset product locator.

    This class implements the `GridSatProductLocator` abstract class
    for the GridSat-GOES/CONUS (Geostationary Operational Environmental
    Satellites - GOES/CONUS) dataset product locator.

    Instances of this class are responsible for generating a list of
    folder paths based on the dataset's directory structure and naming
    conventions, product details, and a specified date range. The
    generated paths cover each year within the date range. Paths to the
    folders containing the initial and final dates are included in the
    list.

    Instances of this class are also responsible for verifying if a
    given filename matches the product filename pattern based on the
    dataset's naming conventions and product specifications, and for
    extracting the corresponding `datetime` information from the
    product's filename.

    The data in the GridSat-GOES/CONUS dataset products comes from GOES
    second generation (GOES-I to GOES-M) series, GOES-8 to GOES-15. The
    dataset provides data for two separate scenes: the entire GOES
    domain (Full Disk) and the CONUS domain (Contiguous United States).
    The scene and origin names are reflected in the product's path and
    filename, as does the product's version. The product's filename
    pattern is as follows:

    'GridSat-<SCENE>.<origin>.<yyyy>.<mm>.<dd>.<HH><MM>.<version>.nc',

    where `<SCENE>` is the scene name in uppercase (e.g. 'CONUS' or
    'GOES'); `<origin>` is the satellite identifier in lowercase (e.g.
    'goes08' to 'goes15'); `<yyyy>` is the gregorian year number;
    `<mm>`, `<dd>`, `<HH>` and <MM> are the month, day, hour and minute,
    respectively, using two digits padded with zeros; and `<version>` is
    the product's version(e.g. 'v01').

    The dataset's files are hosted on the servers of the NOAA's National
    Centers for Environmental Information (NCEI), the product's file
    path pattern is as follows:

    'https://<net-location>/data/gridsat-goes/access/<scene>/<yyyy>/<mm>/',

    where `<net-location>` is 'www.ncei.noaa.gov', and `<scene>` is the
    scene name in lowercase (e.g. 'conus' or 'goes'). `<yyyy>` and
    `<mm>` are, respectively, the gregorian year number and the month
    number using two digits padded with zeros.

    Input is half-hourly data from the GOES second generation satellite
    series with gridded 0.04°x0.04° spatial resolution that spans from
    1994 to 2017. Six total channels are available

    For more information visit the following link and links therein:
    https://www.ncei.noaa.gov/products/satellite/gridded-goes-conus

    Notes
    -----
    No datasource is supported yet.

    Methods
    -------
    get_base_url(datasource: str) -> str:
        Get the base URL for the GridSat-GOES/CONUS imagery dataset's
        products.
    next_time(current_time: datetime) -> datetime
        Get the next time interval. GridSat-GOES/CONUS dataset organises
        the data by month.
    normalise_times(datetime_ini: datetime, datetime_fin: datetime)
        -> tuple[datetime, datetime]
        Normalise the initial and final datetimes.
    truncate_to_month(time: datetime) -> datetime
        Truncate the `datetime` to the current month.

    Caution
    -------
    Members of this class not defined by the `ProductLocator` interface
    are helper methods and can be considered as implementation details,
    even though they are defined as part of the public API. In future
    releases, these methods may be moved to a private scope, suffer
    name changes, or be removed altogether.
    """

    # Satellites in the GOES 2nd generation (GOES-I to GOES-M) series
    # are identified by the following IDs:
    AVAILABLE_ORIGINS: dict[str, str] = {
        f"G{idn:02d}": f"goes{idn:02d}" for idn in range(8, 16)
    }

    # Available scenes for the GridSat-GOES/CONUS imagery dataset
    # products:
    #
    # NOTE: In its strictest sense, “Contiguous United States” refers
    # to the lower 48 states in North America (including the District
    # of Columbia), and “Continental United States” refers to 49 states
    # (including Alaska and the District of Columbia).
    AVAILABLE_SCENES: dict[str, str] = {
        "F": "Full Disk",
        "C": "CONUS (Contiguous United States)",
    }

    # Mapping between scene IDs and dataset group names:
    SCENE_TO_NAME: dict[str, str] = {
        "F": "GOES",
        "C": "CONUS",
    }

    # Base URLs for the available datasources of the GridSat-GOES/CONUS
    # imagery dataset products:
    AVAILABLE_DATASOURCES: dict[str, str] = {
        "HTTP": "https://www.ncei.noaa.gov/data/gridsat-goes/access/",
    }

    # Supported datasources of the GridSat-GOES/CONUS imagery dataset
    # products:
    SUPPORTED_DATASOURCES: set[str] = set(AVAILABLE_DATASOURCES)

    # Available versions of the GridSat-GOES/CONUS imagery dataset
    # products:
    SUPPORTED_VERSIONS: set[str] = {"v01"}

    def __init__(
        self,
        scene: str,
        origins: str | list[str],
        versions: str | list[str] = GOES_PRODUCT_LATEST_VERSION,
    ) -> None:
        """
        Initialise a GridSat-GOES/CONUS imagery dataset product locator.

        Constructs a GridSat-GOES/CONUS imagery dataset product locator
        object.

        Parameters
        ----------
        scene : str
            The scene ID for the dataset. Supported scenes are "F" (Full
            Disk) and "C" (CONUS, Contiguous United States). Due to how
            the GridSat-GOES/CONUS dataset directories are organised,
            only a single scene may be provided.
        origins : str | list[str]
            The origin of the GridSat-GOES/CONUS imagery dataset
            product, namely a satellite identifier, e.g. "G08".
            The origin may be a single origin or a list of origins.
        versions : str | list[str]
            The version of the GridSat-GOES/CONUS imagery dataset
            product; e.g., "v01". The version may be a single version
            or a list of versions. Only the latest version is available
            in the public repository. The default is the latest version
            ("v01").

        Raises
        ------
        ValueError
            If the provided origin, scene, or version is invalid.
        """
        self._validate_scene(scene, self.AVAILABLE_SCENES)

        if isinstance(origins, str):
            origins = [origins]

        if unavailable_origin := set(origins) - set(self.AVAILABLE_ORIGINS):
            available_origins: list[str] = sorted(self.AVAILABLE_ORIGINS)
            raise ValueError(
                f"Invalid origin IDs: {sorted(unavailable_origin)}. "
                f"Available origin IDs: {available_origins}"
            )

        if isinstance(versions, str):
            versions = [versions]

        if unsupported_version := set(versions) - set(self.SUPPORTED_VERSIONS):
            supported_versions: list[str] = sorted(self.SUPPORTED_VERSIONS)
            raise ValueError(
                f"Unsupported versions: {sorted(unsupported_version)}. "
                f"Supported versions: {supported_versions}"
            )

        product_name: str = self.SCENE_TO_NAME[scene]
        data_origin: list[str] = [
            self.AVAILABLE_ORIGINS[orig] for orig in origins
        ]

        goes_path_prefix: str = f"{product_name.lower()}/"

        super().__init__(
            name=product_name,
            origins=data_origin,
            versions=versions,
            file_date_format=GOES_FILE_DATE_FORMAT,
            file_date_pattern=GOES_FILE_DATE_PATTERN,
            file_prefix=GOES_FILE_PREFIX,
            path_date_format=GOES_PATH_DATE_FORMAT,
            path_prefix=goes_path_prefix,
        )

    def get_base_url(self, datasource: str) -> tuple[str, ...]:
        """
        Get the base URL for the GridSat-GOES/CONUS dataset's products.

        This method returns the base URL for the GridSat-GOES/CONUS
        imagery dataset's products. The base URL is used to construct
        the full URL to the dataset's product files.

        Parameters
        ----------
        datasource : str
            The datasource identifier. This parameter is used to
            determine the base URL for the dataset's products. The only
            available datasource is 'NOAA'. No datasource is supported
            yet.

        Returns
        -------
        tuple[str, ...]
            The base URL for the GridSat-GOES/CONUS imagery dataset's
            products based on the requested datasource identifier.

        Raises
        ------
        ValueError
            If the provided datasource is not supported or unavailable.
        """
        if datasource not in self.SUPPORTED_DATASOURCES:
            supported_datasources: list[str] = sorted(
                self.SUPPORTED_DATASOURCES
            )
            raise ValueError(
                f"Unsupported datasource: '{datasource}'. "
                f"Supported datasources: {supported_datasources}"
            )

        return (self.AVAILABLE_DATASOURCES[datasource],)

    def next_time(self, current_time: datetime) -> datetime:
        """
        Get the next time interval.

        Get the next time interval based on the current time interval.

        Parameters
        ----------
        current_time : datetime
            The current time interval.

        Returns
        -------
        datetime
            The next time interval.
        """
        next_month: int = current_time.month + 1
        next_year: int = current_time.year
        if next_month > MONTHS_IN_YEAR:
            next_month = 1
            next_year += 1

        return current_time.replace(year=next_year, month=next_month)

    def normalize_times(
        self, datetime_ini: datetime, datetime_fin: datetime
    ) -> tuple[datetime, datetime]:
        """
        Normalise the initial and final datetimes.

        Normalise the initial and final datetimes to the nearest
        time interval based on the dataset. The initial datetime is
        normalised to the start of the time interval and the final
        datetime is normalised to the end of the time interval.

        Parameters
        ----------
        datetime_ini : datetime
            The initial datetime.
        datetime_fin : datetime
            The final datetime.

        Returns
        -------
        tuple[datetime, datetime]
            A tuple containing the normalised initial and final
            datetimes.
        """
        start_time: datetime = self.truncate_to_month(datetime_ini)
        end_time: datetime = self.truncate_to_month(datetime_fin)

        return start_time, end_time

    @staticmethod
    def truncate_to_month(time: datetime) -> datetime:
        """
        Truncate the `datetime` to the current month.

        The `datetime` is truncated to the beginning of the current
        month.

        Parameters
        ----------
        time : datetime
            The `datetime` to be truncated to the current month.

        Returns
        -------
        datetime
            The `datetime` truncated to the current month.
        """
        return time.replace(day=1, hour=0, minute=0, second=0, microsecond=0)



import json
import os
from pathlib import Path

import numpy as np
from matplotlib import colormaps
from PIL import Image as PImage


def create_animation_file(path, filelist):
    data_json = json.dumps(filelist)
    content_js = f"const animationFrames = {data_json};\n"
    with open(path, "w", encoding="utf-8", newline="\n") as file:
        file.write(content_js)


def create_palette_image(palette_bytes):
    palette_image = PImage.new("P", (1, 1))
    palette_image.putpalette(palette_bytes)

    return palette_image


def create_gif_palette(cmap_names, colors_per_cmap):
    palette_rgb = []
    for i, cmap_name in enumerate(cmap_names):
        cmap = colormaps.get_cmap(cmap_name)
        # Obtener colores RGB normalizados (entre 0 y 1)
        colors_normalized = cmap(np.linspace(0, 1, colors_per_cmap[i]))
        # Convertir a valores enteros entre 0 y 255
        colors_rgb = (colors_normalized[:, :3] * 255).astype(np.uint8)
        palette_rgb.extend(colors_rgb.flatten().tolist())

    return bytes(palette_rgb)


def create_animation(
    media_list, output_path, cmaps=None, fps=None, interval=None, loop=True
):
    # Check for empty path list
    if not media_list:
        print("ERROR: Empty file list!")
        return False

    # Determine the duration of each photogram in ms
    if fps is not None and fps > 0:
        duration_ms = int(1000 / fps)
    elif interval is not None and interval > 0:
        duration_ms = interval
    else:
        duration_ms = 200

    if cmaps is not None:
        n_cmap = len(cmaps)
        colors_size = 256 // n_cmap
        colors_total = colors_size * n_cmap
        colors_per_cmap = [colors_size] * n_cmap
        k = 0
        while colors_total < 256:
            colors_per_cmap[k] += 1
            colors_total += 1
            k = (k + 1) % n_cmap

        # Generar los bytes de la paleta precalculada
        precalculated_palette_bytes = create_gif_palette(
            cmaps, colors_per_cmap
        )

        # Crear una imagen de paleta dummy
        palette_image = create_palette_image(precalculated_palette_bytes)

    # Read the images
    images = []
    for i, file_name in enumerate(media_list):
        if not os.path.exists(file_name):
            print(f"File '{file_name}' does not exist!")
            continue
        try:
            img = PImage.open(file_name).convert("RGB")
            if cmaps is not None:
                # Convertir a modo paleta utilizando la paleta precalculada
                img_p = img.quantize(
                    palette=palette_image, dither=PImage.FLOYDSTEINBERG
                )
                images.append(img_p)
            else:
                images.append(img)
        except Exception as e:
            print(f"File '{file_name}' is corrupted: {e}")

    # Create the animated GIF
    if not images:
        print("ERROR: Empty image list!")
        return False

    # The first frame is the base
    first = images[0]

    # The remaining frames will be added
    remaining = images[1:]

    # Determine the loop parameter value
    # 0=infinite, 1=once
    loop_value = 0 if loop else 1

    generated_by = (
        "Generated by GOES-CORE (https://github.com/wvenialbo/GOES-CORE/)."
    )

    try:
        output_path = Path(output_path)

        if output_path.suffix == ".gif":
            first.save(
                output_path,
                save_all=True,
                append_images=remaining,
                optimize=True,
                # include_color_table=True,
                # palette=0,
                disposal=2,
                loop=loop_value,
                duration=duration_ms,
                comment=generated_by.encode("utf-8"),
            )
        elif output_path.suffix == ".png":
            info = PngInfo()
            info.add_text("Comment", generated_by.encode("utf-8"))
            first.save(
                output_path,
                save_all=True,
                append_images=remaining,
                optimize=True,
                loop=loop_value,
                duration=duration_ms,
                disposal=0,
                blend=0,
                pnginfo=info,
            )
        elif output_path.suffix == ".webp":
            user_comment = b"UNICODE\x00" + generated_by.encode("utf-8")
            exif = PImage.Exif()
            exif[37510] = user_comment
            first.save(
                output_path,
                save_all=True,
                append_images=remaining,
                duration=duration_ms,
                loop=loop_value,
                # lossless=True,
                # quality=100,
                # method=6,
                exif=exif.tobytes(),
            )
        else:
            raise ValueError(f"Unexpected format '{output_path.suffix}'")
        return True

    except Exception as e:
        print(f"ERROR: Unable to create the animation: {e}")
        return False



import numpy as np


def create_radial_grid(shape):
    # Create a grid of pixel indices
    y_coords, x_coords = np.indices((shape), dtype=np.float64)

    # Get the grid extent in cells
    y_extent, x_extent = shape

    # Compute the grid center coordinates
    y_center, x_center = (y_extent - 1) / 2, (x_extent - 1) / 2

    # Compute the distance of each cell to the center
    return np.sqrt((x_coords - x_center) ** 2 + (y_coords - y_center) ** 2)


def radial_profile(data):
    # Compute the distance grid (of each cell to the grid center)
    r_grid = create_radial_grid(data.shape)
    radii = np.round(r_grid).astype(np.int32)

    # Calculate the azimuthal integral for each radii
    azimuthal_integral = np.bincount(radii.ravel(), data.ravel()).astype(
        np.float32
    )

    # Calculate the line integral for each radii
    line_integral = np.bincount(radii.ravel()).astype(np.float32)

    # Mask zero values
    line_integral[line_integral == 0] = np.nan

    # Calulate the azimuthal average for each radii, i.e. for each
    # radii divide the azimuthal integral by the line integral
    profile = azimuthal_integral / line_integral

    # Compute the region radius
    radius = min(data.shape[0] // 2, data.shape[1] // 2)

    return profile, radius


def azimuthal_average_transform(profile, radius, shading="nearest"):
    # Compute the profile extent
    n_profile = len(profile)

    # Compute the side for the output data matrix
    side = 2 * radius

    # Compute the distance grid (of each cell to the grid center)
    r_grid = create_radial_grid((side, side))

    # Compute the azimuthal average transform
    if shading == "nearest":
        # Convert radii to indices
        radii = np.round(r_grid).astype(np.int32)

        # Clip radii to profile extent
        radii = np.clip(radii, 0, n_profile - 1)

        # Transform by lookup table
        return profile[radii]

    if shading == "interp":
        radii = r_grid
        profile_radii = np.arange(n_profile, dtype=np.float32)

        return np.interp(radii, profile_radii, profile)

    raise ValueError("'shading' must be 'interp' oo 'nearest' (default)")


def azimuthal_average_factory(radius, shading="nearest"):
    # Compute the side for the output data matrix
    side = 2 * radius

    # Compute the distance grid (of each cell to the grid center)
    r_grid = create_radial_grid((side, side))

    # Compute the azimuthal average transform
    if shading == "nearest":
        # Convert radii to indices
        radii_nearest = np.round(r_grid).astype(np.int32)

        def transform_nearest(profile):
            # Compute the profile extent
            n_profile = len(profile)

            # Clip radii to profile extent
            radii = np.clip(radii_nearest, 0, n_profile - 1)

            return profile[radii]

        return transform_nearest

    if shading == "interp":
        # Use radii as is
        radii_interp = r_grid

        def transform_interp(profile):
            # Compute the profile extent
            n_profile = len(profile)

            # Create key values
            profile_radii = np.arange(n_profile, dtype=np.float32)

            return np.interp(radii_interp, profile_radii, profile)

        return transform_interp

    raise ValueError("'shading' must be 'interp' or 'nearest' (default)")



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys

# -- Path setup --------------------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#path-setup
src_path = os.path.abspath("../../src/goesdl")
sys.path.insert(0, src_path)


# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "goes-dl"
copyright = "2025, Waldemar Villamayor-Venialbo"
author = "Waldemar Villamayor-Venialbo"
release = "v0.1-rc4"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



"""
Provide local repository mechanisms for managing file operations.

Classes:
    FileRepository: A local repository for managing file operations.
"""

import shutil
from pathlib import Path


class FileRepository:
    """
    A local file repository for managing file operations.

    Provide methods to handle file and directory operations such as
    listing, adding, deleting, moving, and searching files within a
    specified base directory.

    Attributes
    ----------
    base_directory : Path
        The base directory where the repository is created.
    """

    base_directory: Path

    def __init__(self, base_directory: str | Path | None = None) -> None:
        """Initialize the local file repository in a base directory.

        Initialize the repository with a base directory where downloaded
        files will be stored. Create a new directory at the given path
        if it does not exist; any missing parents of this path are
        created as needed.

        Parameters
        ----------
        base_directory : str | Path | None, optional
            The base directory where the repository will be created. If
            not specified, the current working directory is used, by
            default None.

        Raises
        ------
        NotADirectoryError
            If the given path already exists in the file system and is
            not a directory.
        """
        self.base_directory = Path(base_directory or ".")
        if not self.base_directory.exists():
            self.base_directory.mkdir(parents=True)
        elif not self.base_directory.is_dir():
            raise NotADirectoryError(
                f"The path '{base_directory}' is not a directory."
            )

    def add_file(
        self,
        source_path: str | Path,
        target_directory: str | Path = "",
        move: bool | None = None,
    ) -> None:
        """Add a file to the repository by copying it or moving it.

        Copy or move a file from an external location to the specified
        directory within the repository.

        `FileExistsError` is raised if the file already exists in the
        target directory. `FileNotFoundError` if the source file does
        not exist or is not a file. If the target path already exists in
        the file system and is not a directory, `NotADirectoryError` is
        raised.

        Parameters
        ----------
        source_path : str | Path
            The path to the file to be added.
        target_directory : str | Path, optional
            The directory to add the file to, by default "".
        move : bool | None, optional
            Whether to move the file instead of copying it, optional, by
            default False.
        """
        source_path = Path(source_path)
        target_path: Path = self._make_target_path(
            source_path, target_directory
        )
        if move:
            shutil.move(source_path, target_path)
        else:
            shutil.copy2(source_path, target_path)

    def create_directory(self, directory: str | Path) -> None:
        """Create a new directory in the repository.

        Create a new directory inside the repository provided that it
        does not already exist.

        Parameters
        ----------
        directory : str | Path
            The name or relative path of the directory to create.

        Raises
        ------
        FileExistsError
            If the directory already exists.
        """
        dir_path: Path = self.base_directory / directory
        if not dir_path.exists():
            dir_path.mkdir(parents=True)
        else:
            raise FileExistsError(
                f"The directory '{dir_path}' already exists."
            )

    def delete_directory(self, directory: str | Path) -> None:
        """Delete an empty directory from the repository.

        Remove an existing empty directory within the repository.

        Parameters
        ----------
        directory : str | Path
            The name or relative path of the directory to delete.

        Raises
        ------
        NotADirectoryError
            If the directory does not exist.
        OSError
            If the directory is not empty or cannot be deleted.
        """
        dir_path: Path = self.base_directory / directory
        if not dir_path.is_dir():
            raise NotADirectoryError(
                f"The directory '{dir_path}' does not exist "
                "or is not a directory."
            )
        try:
            dir_path.rmdir()
        except OSError as e:
            raise OSError(
                f"The directory '{dir_path}' "
                "is not empty or cannot be deleted."
            ) from e

    def delete_file(
        self, file_name: str | Path, directory: str | Path = ""
    ) -> None:
        """Delete a file from the repository.

        Delete a file by name within a directory in the repository.

        Parameters
        ----------
        file_name : str | Path
            The name or relative path of the file to delete.
        directory : str | Path, optional
            The directory where the file is located, by default "".

        Raises
        ------
        FileNotFoundError
            If the file does not exist or is not a file.
        """
        file_path: Path = self.base_directory / directory / file_name
        if file_path.is_file():
            file_path.unlink()
        else:
            raise FileNotFoundError(
                f"The file '{file_path}' does not exist or is not a file."
            )

    def get_full_path(
        self, file_name: str | Path, directory: str | Path = ""
    ) -> str:
        """Get the full path to a file in the repository.

        Get the full path to a file within a directory in the
        repository.

        Parameters
        ----------
        file_name : str | Path
            The name or relative path of the file.
        directory : str | Path, optional
            The directory where the file is located, by default "".

        Returns
        -------
        str
            The full path to the file.
        """
        return str(self.base_directory / directory / file_name)

    def is_directory(
        self, path_name: str | Path, directory: str | Path = ""
    ) -> bool:
        """Check if a path name from the repository is a directory.

        Check if a path name exists within the repository and is a
        directory.

        Parameters
        ----------
        path_name : str | Path
            The name or relative path name of an object to check.
        directory : str | Path, optional
            The directory where the object is located, by default "".

        Returns
        -------
        bool
            True if the file exists, False otherwise.
        """
        directory_path: Path = self.base_directory / directory / path_name
        return directory_path.is_dir()

    def is_file(
        self, path_name: str | Path, directory: str | Path = ""
    ) -> bool:
        """Check if a path name from the repository is a file.

        Check if a path name exists within a directory in the repository
        and is a file.

        Parameters
        ----------
        path_name : str | Path
            The name or relative path name of an object to check.
        directory : str | Path, optional
            The directory where the object is located, by default "".

        Returns
        -------
        bool
            True if the file exists, False otherwise.
        """
        file_path: Path = self.base_directory / directory / path_name
        return file_path.is_file()

    def list_files(self, directory: str | Path = "") -> list[str]:
        """List all files in the given directory.

        List all files within a given directory or the root directory if
        not specified.

        Parameters
        ----------
        directory : str | Path, optional
            The name or relative path of the directory to list files
            from, by default "".

        Returns
        -------
        list[str]
            A list of file names in the given directory.

        Raises
        ------
        NotADirectoryError
            If the given path is not a directory.
        FileNotFoundError
            If the directory does not exist.
        """
        dir_path: Path = self.base_directory / directory
        if dir_path.is_dir():
            return [item.name for item in dir_path.iterdir() if item.is_file()]
        if dir_path.exists():
            raise NotADirectoryError(
                f"The path '{dir_path}' is not a directory."
            )
        raise FileNotFoundError(f"The directory '{dir_path}' does not exist.")

    def move_file(
        self,
        file_name: str | Path,
        source_directory: str | Path,
        target_directory: str | Path,
    ) -> None:
        """Move a file within the repository.

        Move a file from one directory to another within the repository.

        `FileExistsError` is raised if the file already exists in the
        target directory. `FileNotFoundError` if the source file does
        not exist or is not a file. If the target path already exists in
        the file system and is not a directory, `NotADirectoryError` is
        raised.

        Parameters
        ----------
        file_name : str | Path
            The name or relative path of the file to move.
        source_directory : str | Path
            The directory where the file is currently located.
        target_directory : str | Path
            The directory to move the file to.
        """
        source_path: Path = self.base_directory / source_directory / file_name
        target_path: Path = self._make_target_path(
            source_path, target_directory
        )
        shutil.move(source_path, target_path)

    def path_exists(
        self, path_name: str | Path, directory: str | Path = ""
    ) -> bool:
        """Check if a path exists in the repository.

        Check if a given path exists within the repository.

        Parameters
        ----------
        path_name : str | Path
            The name or relative path name of an object to check.
        directory : str | Path, optional
            The directory where the object is located, by default "".

        Returns
        -------
        bool
            True if the file exists, False otherwise.
        """
        object_path: Path = self.base_directory / directory / path_name
        return object_path.exists()

    def read_file(
        self, file_name: str | Path, directory: str | Path = ""
    ) -> bytes:
        """Read the content of a file in the repository.

        Read the content of a file within a directory in the repository.

        Parameters
        ----------
        file_name : str | Path
            The name or relative path of the file to read.
        directory : str | Path, optional
            The directory where the file is located, by default "".

        Returns
        -------
        bytes
            The content of the file as a byte string.

        Raises
        ------
        FileNotFoundError
            If the file does not exist or is not a file.
        """
        file_path: Path = self.base_directory / directory / file_name
        if file_path.is_file():
            with open(file_path, "rb") as file:
                return file.read()
        raise FileNotFoundError(
            f"The file '{file_path}' does not exist or is not a file."
        )

    def save_file(
        self, content: bytes, file_name: str | Path, directory: str | Path = ""
    ) -> None:
        """Save content to a file in the repository.

        Save the given content to a file within a directory in the
        repository. The relative path will be recreated within the
        repository to mirror the source path.

        Parameters
        ----------
        content : bytes
            The content to save to the file.
        file_name : str | Path
            The name or relative path of the file to save the content.
        directory : str | Path, optional
            The directory to save the file to, by default "".

        Raises
        ------
        FileExistsError
            If the file already exists in the target directory or if the
            given target path already exists in the file system and is
            not a directory (via `mkdir`).
        """
        file_path: Path = self.base_directory / directory / file_name
        file_path.parent.mkdir(parents=True, exist_ok=True)
        if file_path.exists():
            raise FileExistsError(
                f"The file '{file_path}' already exists in the "
                "target directory."
            )
        with open(file_path, "wb") as file:
            file.write(content)

    def search_files(
        self, pattern: str, directory: str | Path = ""
    ) -> list[str]:
        """Search for files that match a pattern.

        Search for files in a given directory within the repository by a
        search pattern; for example, "*.txt" to search for text files.

        Parameters
        ----------
        pattern : str
            The pattern to search for.
        directory : str | Path, optional
            The name or relative path of the directory to search files
            from, by default "".

        Returns
        -------
        list[str]
            A list of file names that match the pattern.

        Raises
        ------
        NotADirectoryError
            If the directory does not exist or is not a directory.
        """
        dir_path: Path = self.base_directory / directory
        if dir_path.is_dir():
            return [item.name for item in dir_path.glob(pattern)]
        raise NotADirectoryError(
            f"The directory '{dir_path}' does not exist "
            "or is not a directory."
        )

    def _make_target_path(
        self, source_path: Path, target_directory: str | Path
    ) -> Path:
        """Create the target directory for adding a file.

        Any missing parents of the target directory path are created as
        needed.

        Parameters
        ----------
        source_path : Path
            The path to the source file.
        target_directory : str | Path
            The directory to add the file to.

        Returns
        -------
        Path
            The target directory path.

        Raises
        ------
        FileExistsError
            If the file already exists in the target directory.
        FileNotFoundError
            If the source file does not exist or is not a file.
        NotADirectoryError
            If the target path already exists in the file system and is
            not a directory.
        """
        target_path: Path = self.base_directory / target_directory
        if not source_path.exists() or not source_path.is_file():
            raise FileNotFoundError(
                f"The file '{source_path}' does not exist or is not a file."
            )
        try:
            target_path.mkdir(parents=True, exist_ok=True)
        except FileExistsError as exc:
            raise NotADirectoryError(
                f"The path '{target_path}' already exists in the "
                "file system and is not a directory."
            ) from exc
        destination: Path = target_path / source_path.name
        if destination.exists():
            raise FileExistsError(
                f"The file '{destination}' already exists in the "
                "target directory."
            )
        return target_path



from numpy import bool_, dtype, float32, float64, int8, ndarray
from numpy.ma import MaskedArray

ArrayBool = ndarray[tuple[int, ...], dtype[bool_]]
ArrayInt8 = ndarray[tuple[int, ...], dtype[int8]]
ArrayFloat32 = ndarray[tuple[int, ...], dtype[float32]]
ArrayFloat64 = ndarray[tuple[int, ...], dtype[float64]]

MaskedFloat32 = MaskedArray[tuple[int, ...], dtype[float32]]



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABIDP: All derived ABI products.
"""

from .locator_nso import GOESProductLocatorABINSO


class GOESProductLocatorABIDP(GOESProductLocatorABINSO):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: All derived ABI products.
    """

    # Available single-band products from the GOES-R Series's ABI
    # instrument:
    AVAILABLE_PRODUCTS: dict[str, str] = {
        "ACHA": "Cloud Top Height (10km for 'F' and 'C', 4km for 'M')",
        "ACHA2KM": "Cloud Top Height (2km since 24 March 2023)",
        "ACHP2KM": "Cloud Top Pressure (2km spatial resolution)",
        "ACHT": "Cloud Top Temperature",
        "ACM": "Clear Sky Mask",
        "ACTP": "Cloud Top Phase",
        "ADP": "Aerosol Detection Product",
        "AICE": "Ice Concentration and Extent",
        "AITA": "Ice Age and Thickness",
        "AOD": "Aerosol Optical Depth",
        "BRF": "Bidirectional Reflectance Factor (Land Surface)",
        "CCL": "Cloud Cover Layers",
        "COD": "Cloud Optical Depth (4km for 'F' and 2km for 'C')",
        "COD2KM": "Cloud Optical Depth (2km since 22 March 2023)",
        "CPS": "Cloud Particle Size",
        "CTP": "Cloud Top Pressure",
        "DSI": "Derived Stability Indices",
        "DSR": "Downward Shortwave Radiation",
        "FDC": "Fire (Hot Spot Characterization)",
        "FSC": "Fractional Snow Cover",
        "LSA": "Land Surface Albedo",
        "LST": "Land Surface Temperature (10km for 'F', 2km for 'C' and 'M')",
        "LST2KM": "Land Surface Temperature (2km since 10 September 2021)",
        "LVMP": "Legacy Vertical Moisture Profile",
        "LVTP": "Legacy Vertical Temperature Profile",
        "RRQPE": "Rainfall Rate (Quantitative Precipitation Estimate)",
        "RSR": "Reflected Shortwave Radiation (Top-Of-Atmosphere)",
        "SST": "Sea Surface Temperature",
        "TPW": "Total Precipitable Water",
        "VAA": "Volcanic Ash (Detection and Height)",
    }

    ONLY_CF_SCENE: set[str] = {
        "AOD",
        "COD",
        "CTP",
        "RSR",
    }

    ONLY_F_SCENE: set[str] = {
        "AICE",
        "AITA",
        "COD2KM",
        "LST2KM",
        "RRQPE",
        "SST",
        "VAA",
    }

    ONLY_FM_SCENE: set[str] = {
        "ACHT",
    }

    ONLY_G16_G17: set[str] = {
        "VAA",
    }

    ONLY_G16_G18: set[str] = {
        "ACHA2KM",
        "ACHP2KM",
        "CCL",
        "FSC",
        "COD2KM",
    }

    CF_SCENE: set[str] = {"C", "F"}
    F_SCENE: set[str] = {"F"}
    FM_SCENE: set[str] = {"F", "M1", "M2"}

    G16_G17_ORIGIN: set[str] = {"G16", "G17"}
    G16_G18_ORIGIN: set[str] = {"G16", "G18"}

    def __init__(self, name: str, scene: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI derived product
        locator object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Raises
        ------
        ValueError
            If the provided scene or origin is invalid.
        """
        self._validate_product(name, self.AVAILABLE_PRODUCTS)

        only_in_segment: list[set[str]] = [
            self.ONLY_CF_SCENE,
            self.ONLY_F_SCENE,
            self.ONLY_FM_SCENE,
        ]
        scene_segment: list[set[str]] = [
            self.CF_SCENE,
            self.F_SCENE,
            self.FM_SCENE,
        ]

        for only, segment in zip(only_in_segment, scene_segment):
            if name in only and scene not in segment:
                raise ValueError(
                    f"Invalid scene '{scene}' "
                    f"for derived ABI product '{name}', "
                    f"supported scenes {sorted(segment)}"
                )

        only_in_segment = [
            self.ONLY_G16_G17,
            self.ONLY_G16_G18,
        ]
        origin_segment: list[set[str]] = [
            self.G16_G17_ORIGIN,
            self.G16_G18_ORIGIN,
        ]

        for only, segment in zip(only_in_segment, origin_segment):
            if name in only and origin not in segment:
                raise ValueError(
                    f"Invalid origin '{origin}' "
                    f"for derived ABI product '{name}', "
                    f"supported origins {sorted(segment)}"
                )

        super().__init__(
            name=name,
            scene=scene,
            origin=origin,
        )



"""
Provide locator for GOES-R Series imagery dataset's ABI products.

Classes:
    - GOESProductLocatorABIDC: All derived ABI products supporting
      channels.
    - GOESProductLocatorDMW: ABI Derived Motion Winds (DMW).
    - GOESProductLocatorDMWV: ABI Derived Motion WV Winds (DMWV).
"""

from .locator_abi import GOESProductLocatorABI


class GOESProductLocatorABIDC(GOESProductLocatorABI):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: All derived ABI products supporting channels.
    """

    DMW_PRODUCT: str = "DMW"
    DMWV_PRODUCT: str = "DMWV"

    AVAILABLE_PRODUCTS: dict[str, str] = {
        DMW_PRODUCT: "Derived Motion Winds",
        DMWV_PRODUCT: "Derived Motion WV Winds",
    }

    M_CHANNELS: set[str] = {"C02", "C07", "C08", "C09", "C10"}
    CF_CHANNELS: set[str] = {"C14"} | M_CHANNELS
    CF_SCENES: set[str] = {"C", "F"}
    WV_CHANNELS: set[str] = {"C08"}

    available_channels: set[str]

    def __init__(
        self, name: str, scene: str, channels: str | list[str], origin: str
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        name : str
            The name of the GOES-R Series imagery dataset ABI product.
            Due to how the dataset directories are organised, only a
            single product can be provided.
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : str | list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C13".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.

        Raises
        ------
        ValueError
            If the provided product name is invalid.
        """
        self._validate_product(name, self.AVAILABLE_PRODUCTS)

        if isinstance(channels, str):
            channels = [channels]

        if not channels:
            raise ValueError(
                f"Derived ABI product '{name}' "
                "does require channel specification"
            )

        self._validate_channels(channels, self.available_channels)

        super().__init__(
            name=name,
            level=self.DEFAULT_PRODUCT_LEVEL,
            scene=scene,
            channels=channels,
            origin=origin,
        )


class GOESProductLocatorDMW(GOESProductLocatorABIDC):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Derived Motion Winds (DMW).
    """

    PRODUCT_NAME_DMW: str = GOESProductLocatorABIDC.DMW_PRODUCT

    def __init__(
        self, scene: str, channels: str | list[str], origin: str
    ) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        channels : str | list[str]
            The list of channels of the GOES-R Series imagery dataset
            ABI product, e.g. "C08" or "C14".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        self.available_channels = (
            self.CF_CHANNELS if scene in self.CF_SCENES else self.M_CHANNELS
        )

        super().__init__(
            name=self.PRODUCT_NAME_DMW,
            scene=scene,
            channels=channels,
            origin=origin,
        )


class GOESProductLocatorDMWV(GOESProductLocatorABIDC):
    """
    Product locator for GOES-R Series imagery dataset's ABI products.

    Instrument: Advanced Baseline Imager (ABI)
    Product: Derived Motion WV Winds (DMWV).
    """

    PRODUCT_NAME_DMWV: str = GOESProductLocatorABIDC.DMWV_PRODUCT

    def __init__(self, scene: str, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset ABI product locator.

        Constructs a GOES-R Series imagery dataset ABI primary product
        locator object.

        Parameters
        ----------
        scene : str
            The scene of the GOES-R Series imagery dataset product, e.g.
            "F" or "C".
        origin : str
            The origin of the GOES-R Series imagery dataset ABI product,
            namely a satellite identifier, e.g. "G16". Due to how the
            dataset directories are organised, only a single origin may
            be provided.
        """
        self.available_channels = self.WV_CHANNELS

        super().__init__(
            name=self.PRODUCT_NAME_DMWV,
            scene=scene,
            channels=list(self.WV_CHANNELS),
            origin=origin,
        )



"""
Provide utility functions for type validation.

Validate a value against a provided type annotation. Include special
handling for NumPy arrays and homogeneous collections.

Functions
---------
validate_value_type(value: Any, annotation: type) -> bool
    Validate the type of a given value against a specified annotation.
"""

import re
from collections.abc import Iterable
from typing import Any, Literal, cast, get_args, get_origin

from numpy import dtype, ndarray


def _validate_ndarray_shape(
    shape: tuple[int, ...], annotation: type | Any
) -> bool:
    # Validate the shape of a NumPy ndarray against an annotation.
    origin = get_origin(annotation)

    if not origin:
        return annotation is Any

    if not issubclass(origin, tuple):
        return False

    specs = get_args(annotation)

    if specs == (int, ...):
        return True

    pattern = ""
    for type_ in specs:
        if type_origin := get_origin(type_):
            if type_origin is not Literal:
                return False
            size = str(get_args(type_)[0])
            if not size.isdigit():
                return False
            lpattern = [f"[{c}]" for c in size]
            pattern += f"{''.join(lpattern)},"
        elif type_ is Ellipsis:
            pattern += r"[\d,]+,"
        elif issubclass(type_, int):
            pattern += r"\d+,"
        else:
            return False

    pattern = pattern.strip(",")
    shape_string = str(shape).strip("(,)").replace(" ", "")

    print(pattern, shape_string)

    return bool(re.fullmatch(pattern, shape_string))


def _validate_ndarray(value: Any, annotation: type) -> bool:
    # Validate a NumPy ndarray against a specified annotation.

    # NOTE: We only expect arrays with scalar numeric content
    # in our project!

    if not isinstance(value, ndarray):
        return False

    # NOTE: According to NumPy's documentation, as of version
    # 2.1.0, the correct way to annotate an 'ndarray' is:
    #
    #        numpy.ndarray[shape, dtype].
    #
    # The 'NDArray' of the 'numpy.typing' module is defined as:
    #
    #  NDArray = numpy.ndarray[tuple[int, ...], dtype].
    #
    # Where 'dtype = numpy.dtype[scalar_type]', 'scalar_type'
    # can be any of 'numpy.int32', 'numpy.float32', etc.
    #
    # Shape typing has been added to NumPy. 'shape' types are
    # covariant and they must be 'tuple[int]', 'tuple[int, int]',
    # 'tuple[int, ...]', etc. 'Any' is accepted for backward
    # compatibility. This allows things like 'tuple[Literal[3]]'
    # to be valid shape type hinting. 'Any' can be used for
    # typing arrays with a given 'dtype' and unspecified shape.

    if dtype[value.dtype] != get_args(annotation)[1]:
        return False

    return _validate_ndarray_shape(value.shape, get_args(annotation)[0])


def validate_type(value: Any, annotation: type) -> bool:
    """
    Validate the type of a given value against a specified annotation.

    Validate the type of a given value against a specified annotation,
    including special handling for NumPy arrays and homogeneous
    collections.

    Parameters
    ----------
    value : Any
        The value to be validated.
    annotation : type
        The annotation to be validated against.

    Returns
    -------
    bool
        True if the value matches the annotation, False otherwise.
    """
    if isinstance(annotation, type(Any)):
        return True

    origin: type | None = get_origin(annotation)

    if not origin:
        return isinstance(value, annotation)

    if not isinstance(value, origin):
        return False

    if isinstance(value, ndarray):
        return _validate_ndarray(value, annotation)

    # NOTE: We only expect collections containing only
    # homogeneous scalar content in our project!
    # No nested structures are expected.

    # If it reached here, we are witnessing a sequence or
    # iterable (set or dictionary). We could collect the
    # types of each element and check if they match the
    # types in the annotation. However, depending on the
    # size of these collections, this would be excessive.
    # Therefore, we just ignore the content. The case of
    # an 'ndarray' is different, as it is homogeneous and
    # its dimensions and element type are annotated through
    # the 'shape' and 'dtype' attributes, respectively.

    arguments: tuple[type, ...] = get_args(annotation)

    if len(arguments) > 1 or get_origin(arguments[0]):
        return False

    # We know it is an iterable, so we can cast it.
    value = cast(Iterable[Any], value)

    return all(isinstance(item, arguments[0]) for item in value)



"""
Provide the EnhacementTable class and related utilities for
creating and manipulating color enhancement tables.
"""

from collections.abc import Sequence
from pathlib import Path
from typing import Literal

from .palette import EnhacementPalette
from .shared import ColorEntry, ColorSegment, DomainData, PaletteData, RGBValue
from .stretching import EnhacementStretching
from .utility import interp, interpx

ColorKey = Literal["red", "green", "blue", "alpha"]
ColorStock = dict[str, RGBValue]
PaletteTable = dict[ColorKey, Sequence[ColorSegment]]


class EnhacementTable:
    """
    Represent a color enhancement table.

    This class provides methods to create an enhancement table from
    stretching and palette data, and to extract sub-palettes.

    Attributes
    ----------
    name : str
        The name or identifier of the enhancement table.
    stock : ColorStock
        The stock color values.
    table : PaletteData
        The palette data containing color segments.
    palette : PaletteTable
        The palette table containing color entries.

    Properties
    ----------
    domain : DomainData
        The input domain containing the minimum and maximum input values.
    extent : DomainData
        The palette domain containing the minimum and maximum values.

    Methods
    -------
    extract(vmin, vmax)
        Extract a sub-palette from the color enhancement table.
    from_file(stretching_path, palette_path)
        Create an EnhacementTable instance from files.
    reverse()
        Reverse the enhancement color table.
    """

    name: str
    stock: ColorStock
    color_table: PaletteTable
    color_data: PaletteData
    palette: EnhacementPalette
    stretching: EnhacementStretching

    def __init__(
        self, palette: EnhacementPalette, stretching: EnhacementStretching
    ) -> None:
        self.name = palette.name
        if stretching.name:
            self.name = f"{self.name} > {stretching.name}"

        self.stock = self._make_stock(palette)
        self.color_data = self._stretch_palette(stretching, palette)
        self.color_table = self._make_colortable(self.color_data)

        self.stretching = stretching
        self.palette = palette

    def extract(self, vmin: float, vmax: float) -> PaletteTable:
        """
        Extract a sub-palette from the color enhancement table.

        Parameters
        ----------
        vmin : float
            The minimum value for the sub-palette.
        vmax : float
            The maximum value for the sub-palette.

        Returns
        -------
        PaletteData
            The extracted sub-palette data.
        """
        tmin, tmax = self.stretching.domain
        trange = tmax - tmin

        vmin = (vmin - tmin) / trange
        vmax = (vmax - tmin) / trange
        cmin = self._interp_color(vmin, self.color_data)
        cmax = self._interp_color(vmax, self.color_data)

        sub_palette: PaletteData = [cmin]
        for entry in self.color_data:
            if entry[0] <= cmin[0] or entry[0] >= cmax[0]:
                continue
            sub_palette.append(entry)

        sub_palette.append(cmax)

        self._normalize_palette(vmin, vmax, sub_palette)

        return self._make_colortable(sub_palette)

    @classmethod
    def from_file(
        cls, palette_path: str | Path, stretching_path: str | Path = ""
    ) -> "EnhacementTable":
        """
        Create an EnhacementTable instance from files.

        Parameters
        ----------
        palette_path : str or Path
            Path to the file containing the enhancement palette data.
        stretching_path : str or Path
            Path to the file containing the enhancement stretching data.

        Returns
        -------
        EnhacementTable
            An instance of the EnhacementTable class.
        """
        palette = EnhacementPalette.from_file(palette_path)

        stretching = None
        if stretching_path:
            stretching = EnhacementStretching.from_file(stretching_path)

        if not stretching:
            table = [(x, x) for x, _, _, _ in palette.table]
            scale_domain, palette_extent = palette.extent, palette.extent
            stretching = EnhacementStretching(
                "", table, scale_domain, palette_extent
            )

        return cls(palette, stretching)

    @staticmethod
    def _interp_color(x: float, color_data: PaletteData) -> ColorEntry:
        x_pal, b_pal, g_pal, r_pal = zip(*color_data)

        # Linear interpolation between the two points
        b, g, r = (
            interp(x, x_pal, b_pal),
            interp(x, x_pal, g_pal),
            interp(x, x_pal, r_pal),
        )

        return x, b, g, r

    @staticmethod
    def _make_colortable(table: PaletteData) -> PaletteTable:
        blue: list[ColorSegment] = []
        green: list[ColorSegment] = []
        red: list[ColorSegment] = []

        _, bp, gp, rp = table[0]
        for i, (x, b, g, r) in enumerate(table):
            if i % 2 == 0 and i > 0:
                blue.append((x, bp, b))
                green.append((x, gp, g))
                red.append((x, rp, r))
            else:
                blue.append((x, b, b))
                green.append((x, g, g))
                red.append((x, r, r))
                bp, gp, rp = b, g, r

        return {
            "red": red,
            "green": green,
            "blue": blue,
        }

    @staticmethod
    def _make_stock(palette: EnhacementPalette) -> ColorStock:
        bg, fg, nn = palette.stock
        return {"background": bg, "foreground": fg, "nan": nn}

    @staticmethod
    def _normalize_palette(
        vmin: float, vmax: float, sub_palette: PaletteData
    ) -> None:
        vrange = vmax - vmin
        for i, entry in enumerate(sub_palette):
            x, b, g, r = entry
            x = (x - vmin) / vrange
            sub_palette[i] = x, b, g, r

    @staticmethod
    def _stretch_palette(
        stretching: EnhacementStretching, palette: EnhacementPalette
    ) -> PaletteData:
        x_stretch: tuple[float]
        y_stretch: tuple[float]
        y_stretch, x_stretch = zip(*stretching.table)

        linearized_table: PaletteData = []
        for x, b, g, r in palette.table:
            y = interpx(x, x_stretch, y_stretch)
            linearized_table.append((y, b, g, r))

        return linearized_table

    @property
    def domain(self) -> DomainData:
        return self.stretching.domain

    @property
    def extent(self) -> DomainData:
        return self.palette.extent



import unittest
from datetime import datetime, timedelta, timezone

from goesdl.gridsat import GridSatProductLocatorB1


class TestPassed(Exception):
    """
    Exception raised when a test has passed.
    """


class TestGridSatProductLocatorB1(unittest.TestCase):
    # This set of tests covers the initialisation of an object of
    # the `GridSatProductLocatorB1` class, including the handling of
    # invalid parameters. It also tests the `ProductLocator` interface
    # implemented by the `GridSatProductLocatorB1` and parent classes.

    def setUp(self) -> None:
        # Initialise the locator with all supported versions.
        self.locator: GridSatProductLocatorB1 = GridSatProductLocatorB1(
            versions=self._supported_versions()
        )
        self.longMessage = True

    def setUpLocator(self, version: str = "") -> None:
        version = version or self._supported_versions()[0]
        self.locator: GridSatProductLocatorB1 = GridSatProductLocatorB1(
            versions=version
        )

    # ------------------------------------------------------------------
    # A minimal subset of tests are designed to reach 100% code coverage
    # of the `GridSatProductLocatorB1` class hierarchy. The remaining
    # complementary tests help to reach 100% test coverage.
    #
    # Note: For completeness, `match()` is tested against valid filename
    # format with an unknown version, and against a valid filename with
    # an invalid date.
    #
    # Although date conversion is performed under the hood by the
    # framework and `get_datetime()` does not perform any checking for
    # invalid dates, it is tested here against a valid filename with an
    # invalid date to ensure that the method is consistent. It is also
    # tested against a `datetime` object set in a non-UTC timezone.

    def test_init_default_version(self) -> None:
        # Initialise the locator with the default version.
        with self.assertRaises(TestPassed):
            GridSatProductLocatorB1()
            raise TestPassed()

    def test_init_supported_versions(self) -> None:
        for version in self._supported_versions():
            with self.subTest(version=version):
                with self.assertRaises(TestPassed):
                    GridSatProductLocatorB1(versions=version)
                    raise TestPassed()

    def test_init_unsupported_versions(self) -> None:
        for version in self._unsupported_versions():
            with self.subTest(version=version):
                with self.assertRaises(ValueError):
                    GridSatProductLocatorB1(versions=version)

    def test_get_base_url_supported_datasources(self) -> None:
        SUPPORTED_URLS: dict[str, str] = self._supported_urls()
        for datasource_id, expected_url in SUPPORTED_URLS.items():
            returned_url: str = self.locator.get_base_url(
                datasource=datasource_id
            )
            with self.subTest(datasource=datasource_id):
                self.assertEqual(returned_url, expected_url)

    def test_get_base_url_unsupported_datasources(self) -> None:
        for datasource_id in self._unsupported_datasources():
            with self.subTest(datasource=datasource_id):
                with self.assertRaises(ValueError):
                    self.locator.get_base_url(datasource=datasource_id)

    def test_get_datetime_valid_date_utc(self) -> None:
        # Expected `datetime` object set in UTC timezone.
        TZ_UTC: timezone = timezone.utc
        self._test_get_datetime_valid_date(14, TZ_UTC)

    def test_get_datetime_valid_date_non_utc(self) -> None:
        # Expected `datetime` object set in different non-UTC timezone.
        TZ_PYT: timezone = timezone(timedelta(hours=-4.0))
        self._test_get_datetime_valid_date(10, TZ_PYT)

    def test_get_datetime_bad_date_format(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                with self.assertRaises(ValueError):
                    self.locator.get_datetime(invalid_filename)

    def test_get_datetime_incomplete_date(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                with self.assertRaises(ValueError):
                    self.locator.get_datetime(invalid_filename)

    def test_get_datetime_invalid_date(self) -> None:
        # Valid filename with invalid date, e.g. April 31.
        VALID_FILENAME_INVALID_DATE: str = self._filename(month=4, day=31)
        with self.assertRaises(ValueError):
            self.locator.get_datetime(VALID_FILENAME_INVALID_DATE)

    def test_match_valid_filename(self) -> None:
        VALID_FILENAME: str = self._filename()
        self.assertTrue(self.locator.match(VALID_FILENAME))

    def test_match_misspelled_prefix(self) -> None:
        INVALID_FILENAME: str = self._filename(prefix="GRIDCAT")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_misspelled_name(self) -> None:
        INVALID_FILENAME: str = self._filename(name="B2")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_date_format(self) -> None:
        for invalid_filename in self._bad_date_format():
            with self.subTest(filename=invalid_filename):
                self.assertFalse(self.locator.match(invalid_filename))

    def test_match_incomplete_date(self) -> None:
        for invalid_filename in self._incomplete_dates():
            with self.subTest(filename=invalid_filename):
                self.assertFalse(self.locator.match(invalid_filename))

    def test_match_invalid_date(self) -> None:
        # Valid filename format with invalid dates are not caught
        # by `match()` because the date format is correct and do not
        # violates the filename pattern, although it should be caught
        # by the `get_datetime()` method.
        #
        # See `test_get_datetime_invalid_date()`.
        VALID_FILENAME_INVALID_DATE: str = self._filename(month=4, day=31)
        self.assertTrue(self.locator.match(VALID_FILENAME_INVALID_DATE))

    def test_match_supported_versions(self) -> None:
        for version in self._supported_versions():
            self.setUpLocator(version=version)
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_versions(self) -> None:
        self.setUpLocator(version=self._supported_versions()[0])
        for version in self._unsupported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertFalse(self.locator.match(filename))

    def test_match_supported_multi_versions(self) -> None:
        for version in self._supported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertTrue(self.locator.match(filename))

    def test_match_unsupported_multi_versions(self) -> None:
        for version in self._unsupported_versions():
            filename: str = self._filename(version=version)
            with self.subTest(version=version):
                self.assertFalse(self.locator.match(filename))

    def test_match_wrong_suffix(self) -> None:
        INVALID_FILENAME: str = self._filename(suffix=".txt")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_name_separator(self) -> None:
        INVALID_FILENAME: str = self._filename().replace("-B1", "_B1")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_date_separator(self) -> None:
        FILENAME: str = self._filename(year=1999, month=12, day=31, hour=23)
        invalid_filename: str = FILENAME.replace(".1999", "_1999")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Year separator"
        )
        invalid_filename = FILENAME.replace(".12", "_12")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Month separator"
        )
        invalid_filename = FILENAME.replace(".31", "_31")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Day separator"
        )
        invalid_filename = FILENAME.replace(".23", "_23")
        self.assertFalse(
            self.locator.match(invalid_filename), msg="Hour separator"
        )

    def test_match_bad_version_separator(self) -> None:
        INVALID_FILENAME: str = self._filename().replace(".v02r01", "_v02r01")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_suffix_separator(self) -> None:
        INVALID_FILENAME: str = self._filename(suffix="_nc")
        self.assertFalse(self.locator.match(INVALID_FILENAME))

    def test_match_bad_version_prefix(self) -> None:
        FILENAME: str = self._filename()
        invalid_filename: str = FILENAME.replace(".v02r01", ".b02r01")
        self.assertFalse(self.locator.match(invalid_filename))
        invalid_filename: str = FILENAME.replace(".v02r01", ".v02s01")
        self.assertFalse(self.locator.match(invalid_filename))

    def test_get_paths(self) -> None:
        TIME_1: datetime = datetime(1970, 8, 23, 0)
        TIME_2: datetime = datetime(2020, 8, 23, 14)
        RETURNED_PATHS: list[str] = self.locator.get_paths(TIME_1, TIME_2)
        EXPECTED_PATHS: list[str] = [
            self._path(year) for year in range(1970, 2021)
        ]
        self.assertEqual(RETURNED_PATHS, EXPECTED_PATHS)

    # ------------------------------------------------------------------

    def _test_get_datetime_valid_date(self, hour: int, tz: timezone) -> None:
        VALID_FILENAME: str = self._filename()
        EXPECTED_DATETIME: datetime = datetime(2020, 8, 23, hour, tzinfo=tz)
        RETURNED_DATETIME: datetime = self.locator.get_datetime(VALID_FILENAME)
        self.assertEqual(RETURNED_DATETIME, EXPECTED_DATETIME)

    # ------------------------------------------------------------------

    @staticmethod
    def _filename(
        prefix: str = "GRIDSAT",
        name: str = "B1",
        year: int = 2020,
        month: int = 8,
        day: int = 23,
        hour: int = 14,
        version: str = "v02r01",
        suffix: str = ".nc",
    ) -> str:
        # By default returns a valid filename for the GridSat-B1 imagery
        # dataset products.
        return (
            f"{prefix}-{name}."
            f"{year:04d}.{month:02d}.{day:02d}.{hour:02d}."
            f"{version}{suffix}"
        )

    @classmethod
    def _bad_date_format(cls) -> list[str]:
        BAD_YEAR: str = cls._filename(year=99999)
        BAD_MONTH: str = cls._filename(month=999)
        BAD_DAY: str = cls._filename(day=999)
        BAD_HOUR: str = cls._filename(hour=999)
        return [BAD_YEAR, BAD_MONTH, BAD_DAY, BAD_HOUR]

    @classmethod
    def _incomplete_dates(cls) -> list[str]:
        NO_YEAR: str = cls._filename(year=9999).replace(".9999", "")
        NO_MONTH: str = cls._filename(month=99).replace(".99", "")
        NO_DAY: str = cls._filename(day=99).replace(".99", "")
        NO_HOUR: str = cls._filename(hour=99).replace(".99", "")
        return [NO_YEAR, NO_MONTH, NO_DAY, NO_HOUR]

    @staticmethod
    def _path(year: int) -> str:
        return f"{year:04d}/"

    # ------------------------------------------------------------------

    @staticmethod
    def _available_datasources() -> dict[str, str]:
        return {
            "AWS": "s3://noaa-cdr-gridsat-b1-pds/data/",
            "Azure": "https://noaa-cdr-gridsat-b1.blob.core.windows.net/data/",
            "GCP": "gs://noaa-cdr-gridsat-b1/data/",
            "NOAA": "https://www.ncei.noaa.gov/data/geostationary-ir-"
            "channel-brightness-temperature-gridsat-b1/access/",
        }

    @staticmethod
    def _supported_datasources() -> list[str]:
        return ["AWS"]

    @classmethod
    def _supported_urls(cls) -> dict[str, str]:
        AVAILABLE_DATASOURCES: dict[str, str] = cls._available_datasources()
        SUPPORTED_DATASOURCES: list[str] = cls._supported_datasources()
        return {
            datasource_id: url
            for datasource_id, url in AVAILABLE_DATASOURCES.items()
            if datasource_id in SUPPORTED_DATASOURCES
        }

    @classmethod
    def _unsupported_datasources(cls) -> list[str]:
        AVAILABLE_DATASOURCES: dict[str, str] = cls._available_datasources()
        SUPPORTED_DATASOURCES: list[str] = cls._supported_datasources()
        return [
            datasource_id
            for datasource_id in AVAILABLE_DATASOURCES.keys()
            if datasource_id not in SUPPORTED_DATASOURCES
        ]

    # ------------------------------------------------------------------

    @staticmethod
    def _supported_versions() -> list[str]:
        return ["v02r01"]

    @classmethod
    def _unsupported_versions(cls) -> list[str]:
        AVAILABLE_VERSIONS: set[str] = {
            f"v{v:02}r{r:02}" for v in range(1, 10) for r in range(1, 6)
        }
        SUPPORTED_VERSIONS: list[str] = cls._supported_versions()
        return [
            version
            for version in AVAILABLE_VERSIONS
            if version not in SUPPORTED_VERSIONS
        ]



"""
Provide the EnhacementStretching class and related functions
for handling color enhancement stretching mappings.

Classes
-------
EnhacementStretching
    Represent a color enhancement stretching mapping.
"""

from pathlib import Path
from typing import TextIO

from .constants import UNNAMED_TABLE
from .shared import DomainData, StretchingTable

ST_SIGNATURE = "ST TABLE"
ST_KEYWORD = (
    "ST",
    "TABLE",
    "Brightness",
    "Temperature",
    "Color",
    "Index",
    "---",
    "------",
)


class EnhacementStretching:
    """
    Represent a color enhancement stretching mapping.

    Provide methods to create an enhancement stretching instance from a
    file and to normalize the enhancement stretching table.

    Attributes
    ----------
    name : str
        The name of the enhancement stretching table.
    table : StretchingTable
        A list of tuples representing the enhancement stretching table.
        The stretching table containing pairs of (x, y) values.
    domain : tuple[DomainData, DomainData]
        The domain data for the x-axis and y-axis containing the minimum
        and maximum values.

    Methods
    -------
    create_file(path, name, table, domain_x, domain_y)
        Create a file with the enhancement stretching table.
    from_file(path)
        Load a enhancement stretching table specification and create an
        EnhacementStretching instance.
    reverse()
        Reverse the enhancement stretching table.
    save_to_file(path, name)
        Save the enhancement stretching table to a file.
    """

    name: str
    domain: DomainData
    extent: DomainData
    table: StretchingTable

    def __init__(
        self,
        name: str,
        table: StretchingTable,
        scale_domain: DomainData,
        palette_extent: DomainData,
    ) -> None:
        self.name = name
        self.domain = scale_domain
        self.extent = palette_extent
        self.table = table

    @classmethod
    def create_file(
        cls,
        path: str | Path,
        name: str,
        table: StretchingTable,
        domain_x: DomainData,
        domain_y: DomainData,
    ) -> None:
        """
        Create a file with the enhancement stretching table.

        Parameters
        ----------
        path : str or Path
            Path to the file where the enhancement stretching table will
            be saved.
        name : str
            The name of the enhancement stretching table.
        table : StretchingTable
            A list of tuples representing the enhancement stretching
            table.
        domain_x : DomainData
            The domain data for the x-axis containing the minimum and
            maximum values.
        domain_y : DomainData
            The domain data for the y-axis containing the minimum and
            maximum values.
        """
        lines: list[str] = []

        cls._add_stretching_table_header(lines, name)
        cls._create_stretching_table(lines, table, domain_x, domain_y)

        with open(path, "w", encoding="utf-8", newline="\n") as file:
            cls._write_stretching_table_file(file, lines)

    @classmethod
    def from_file(cls, path: str | Path) -> "EnhacementStretching":
        """
        Create an EnhacementStretching instance from a file.

        Parameters
        ----------
        path : str or Path
            Path to the file containing the enhancement stretching data.

        Returns
        -------
        EnhacementStretching
            An instance of the EnhacementStretching class.
        """
        with open(path, "r", encoding="utf-8") as file:
            lines = file.readlines()

        table, name = cls._parse_table(lines)

        scale_domain, palette_extent = cls._normalize_table(table)

        return cls(name, table, scale_domain, palette_extent)

    def save_to_file(self, path: str | Path, name: str = "") -> None:
        """
        Save the enhancement stretching table to a file.

        Parameters
        ----------
        path : str or Path
            Path to the file where the enhancement stretching table will
            be saved.
        name : str
            The name of the enhancement stretching table. Override the
            actual name if provided.
        """
        if not name and self.name != UNNAMED_TABLE:
            name = self.name

        self.create_file(path, name, self.table, self.domain, self.extent)

    @staticmethod
    def _add_stretching_table_header(lines: list[str], name: str) -> None:
        lines.extend(
            (
                f"{ST_SIGNATURE} {name.upper()}",
                f" {ST_KEYWORD[3]} {ST_KEYWORD[4]}",
                f"{ST_KEYWORD[2]}  {ST_KEYWORD[5]}",
                f"{ST_KEYWORD[6]:>8}{ST_KEYWORD[6]:>9}",
            )
        )

    @staticmethod
    def _create_stretching_table(
        lines: list[str],
        table: StretchingTable,
        domain_x: DomainData,
        domain_y: DomainData,
    ) -> None:
        x_min, x_max = domain_x
        y_min, y_max = domain_y

        x_range = x_max - x_min
        y_range = y_max - y_min

        for x_i, y_i in table:
            x = round(x_min + x_i * x_range)
            y = round(y_min + y_i * y_range)

            lines.append(f"{x:>8}{y:>9}")

    @classmethod
    def _normalize_table(
        cls, table: StretchingTable
    ) -> tuple[DomainData, DomainData]:
        x_min, y_min = table[0]
        x_max, y_max = table[-1]
        x_scl, y_scl = x_max - x_min, y_max - y_min

        for i, (x, y) in enumerate(table):
            table[i] = ((x - x_min) / x_scl, (y - y_min) / y_scl)

        return (x_min, x_max), (y_min, y_max)

    @staticmethod
    def _parse_table(lines: list[str]) -> tuple[StretchingTable, str]:
        if ST_SIGNATURE not in lines[0]:
            raise ValueError("Invalid stretching table")

        name = UNNAMED_TABLE
        if len(lines[0]) > len(ST_SIGNATURE):
            name = lines[0][len(ST_SIGNATURE) + 1 :]
            name = name.strip()

        table: StretchingTable = []
        for line in lines:
            ls = line.split()
            if ls[0] in ST_KEYWORD:
                continue
            x, y = map(float, ls)
            table.append((x, y))
        return table, name

    @staticmethod
    def _write_stretching_table_file(file: TextIO, lines: list[str]) -> None:
        for line in lines:
            file.write(f"{line:<85}\n")



"""
Experimental code for potential future integration; API may change.

This subpackage houses code that is currently under testing or
development for potential inclusion in the main package in the future.

**Important:** Interfaces and functionality within this subpackage are
subject to change or removal without prior notice in future versions.
Users should be aware that the code found here is not considered stable
and might not adhere to the same standards as the main package.

Furthermore, code within this 'experimental' subpackage should **not**
be subjected to static analysis tools such as linters, as it may contain
work-in-progress elements or deviations from established style
guidelines.

Finally, please note that the code in this subpackage might not be fully
annotated with type hints.
"""



"""
Calculate latitude and longitude grids data using cartopy.

Notes
-----
See GOES-R Product User Guide (PUG) Volume 5 (L2 products) Section 4.2.8
for details & example of calculations.

Functions
---------
calculate_latlon_grid_cartopy
    Calculate latitude and longitude grids using the cartopy package.
"""

from numpy import float32, meshgrid

from ..utils.array import ArrayFloat32, ArrayFloat64
from .helpers import make_consistent
from .geostationary_parameters import GeostationaryParameters


def geos_to_latlon_grid_cartopy(
    projection_info: GeostationaryParameters,
) -> tuple[ArrayFloat32, ArrayFloat32]:
    """
    Calculate latitude and longitude grids using the cartopy package.

    Calculate latitude and longitude from GOES ABI fixed grid projection
    data. GOES ABI fixed grid projection is a map projection relative to
    the GOES satellite.

    Units: latitude in °N (°S < 0), longitude in °E (°W < 0)

    Parameters:
    -----------
    projection_info : ProjectionParameters
        Object containing the satellite's projection information and
        GOES ABI fixed grid data.

    Returns:
    --------
    tuple[ArrayFloat32, ArrayFloat32]
        A tuple containing the latitude and longitude grid data.
    """
    try:
        import cartopy.crs as ccrs
    except ImportError as error:
        raise ImportError(
            "The 'cartopy' package is required for this functionality."
        ) from error

    globe_geos = ccrs.Globe(
        ellipse=None,
        semimajor_axis=projection_info.globe.semi_major_axis,
        semiminor_axis=projection_info.globe.semi_minor_axis,
        inverse_flattening=projection_info.globe.inverse_flattening,
    )

    geos_proj = ccrs.Geostationary(
        satellite_height=projection_info.orbit.perspective_point_height,
        central_longitude=projection_info.orbit.longitude_of_projection_origin,
        sweep_axis=projection_info.orbit.sweep_angle_axis,
        globe=globe_geos,
    )

    plate_carree_proj = ccrs.PlateCarree(globe=globe_geos)

    x_m: ArrayFloat64
    y_m: ArrayFloat64
    x_m, y_m = meshgrid(projection_info.x_m, projection_info.y_m)

    points = plate_carree_proj.transform_points(geos_proj, x_m, y_m)

    abi_lon: ArrayFloat64 = points[..., 0]
    abi_lat: ArrayFloat64 = points[..., 1]

    abi_lon, abi_lat = make_consistent(abi_lon, abi_lat)

    return abi_lat.astype(float32), abi_lon.astype(float32)



"""
Provide utility classes and functions for the GOES-DL project.

Classes:
    FileRepository: A class for handling file operations.
    RequestHeaders: A class for managing HTTP request headers.
    url: A utility function for URL manipulations.
"""

from .file_repository import FileRepository
from .headers import RequestHeaders
from .url import URL

__all__ = ["FileRepository", "RequestHeaders", "URL"]



"""
Provide functionality for reading fragments NetCDF datasets.

NetCDF (Network Common Data Form) is a file format and set of software
libraries for storing and sharing scientific data. It's designed to make
it easy to access array-oriented data.

This package provides classes to extract and represent information from
NetCDF data files. The classes dynamically set attributes based on the
annotations defined in the class and the corresponding attributes or
variables present in the provided NetCDF data object.
"""

from . import fields
from .class_help import HasStrHelp
from .fields import (
    attribute,
    computed,
    data,
    dimension,
    field,
    scalar,
    variable,
)
from .view import DatasetView

__all__ = [
    "attribute",
    "computed",
    "data",
    "DatasetView",
    "dimension",
    "field",
    "fields",
    "HasStrHelp",
    "scalar",
    "variable",
]



from .shared import DomainData

MIN_TICKS = 5


class ColorbarTicks:

    cticks: list[float]

    def __init__(self, extent: DomainData, nticks: int, step: int) -> None:
        """
        Create a list of colorbar ticks for the given temperature range.

        Creates a list of ticks evenly spaced between tmin and tmax,
        with a minimum step of 5.

        Parameters
        ----------
        extent: DomainData
            Tuple with minimum and maximum temperature.
        nticks : int
            Maximum number of ticks.
        step : int, optional
            Tick step, by default 0 (automatic).

        Returns
        -------
        list[float]
            List of colorbar ticks.
        """
        tmin, tmax = extent

        if step < MIN_TICKS:
            step = self._find_tick_step(
                tmin, tmax, max_ticks=nticks, min_step=MIN_TICKS
            )

        cbmin = self._find_tick_min(tmin, step)
        cbmax = self._find_tick_max(tmax, step) + 1

        self.cticks = [float(tick) for tick in range(cbmin, cbmax, step)]

    @staticmethod
    def _find_tick_max(vmax: float, step: int) -> int:
        """
        Finds the last multiple of `step` that is less or equal to `vmax`.

        Parameters
        ----------
        vmax : float
            The number from which the last multiple less or equal will be
            searched.
        step : int
            The number of which we want to find the multiples.

        Returns
        -------
        int
            The last multiple of `step` that is less or equal to `vmax`.
        """
        cociente = vmax // step
        return int(cociente) * step

    @staticmethod
    def _find_tick_min(vmin: float, step: int) -> int:
        """
        Finds the first multiple of `step` that is greater or equal to
        `vmin`.

        Parameters
        ----------
        vmin : float
            The number from which the first multiple greater or equal will
            be searched.
        step : int
            The number of which we want to find the multiples.

        Returns
        -------
        int
            The first multiple of `step` that is greater or equal to `vmin`.
        """
        quotient, remainder = divmod(vmin, step)
        return int(quotient + 1 if remainder else quotient) * step

    @classmethod
    def _find_tick_step(
        cls, tmin: float, tmax: float, max_ticks: int, min_step: int
    ) -> int:
        """
        Finds the step that divides the range [tmin, tmax] into `ticks`
        intervals.

        Parameters
        ----------
        tmin : float
            The minimum value of the range.
        tmax : float
            The maximum value of the range.
        max_ticks : int, optional
            The maximum number of intervals to divide the range.
        min_step : int, optional
            The minimum tick step to consider. Larger steps will be
            multiple of this value.

        Returns
        -------
        float
            The step that divides the range [tmin, tmax] into up
            to `max_ticks` intervals.
        """
        return cls._find_tick_min((tmax - tmin) / max_ticks, min_step)



def standard_name_to_label(standard_name):
    words = standard_name.split("_")
    labels = []
    acronyms = ["TOA", "IR", "UV", "SST"]

    for word in words:
        if palabra.upper() in acronyms:
            labels.append(word.upper())
        else:
            labels.append(word.capitalize())

    return " ".join(labels)



"""
Provide locator for GOES-R Series imagery dataset's GLM products.

Classes:
    - GOESProductLocatorLCFA: GLM Lightning Cluster-Filter Algorithm
      (LCFA).
"""

from .locator_glm import GOESProductLocatorGLM


class GOESProductLocatorLCFA(GOESProductLocatorGLM):
    """
    Product locator for GOES-R Series imagery dataset's GLM products.

    Instrument: Geostationary Lightning Mapper (GLM).
    Product: Lightning Cluster-Filter Algorithm (LCFA).
    """

    PRODUCT_NAME: str = "LCFA"

    def __init__(self, origin: str) -> None:
        """
        Initialise a GOES-R Series imagery dataset LCFA product locator.

        Constructs a GOES-R Series imagery dataset LCFA product locator
        object.

        Parameters
        ----------
        origin : str
            The origin of the GOES-R Series dataset LCFA product, namely
            a satellite identifier, e.g. "G16". Due to how the dataset
            directories are organised, only a single origin may be
            provided.
        """
        super().__init__(name=self.PRODUCT_NAME, origin=origin)



import colorsys
import math

from .clr_table import clr_utility
from .constants import (
    BRG_MAX,
    CM_CMYK,
    CM_GRAY,
    CM_HSV,
    CM_RGB,
    CMYK_MAX,
    HSV_MAX,
    HUE_MAX,
    UNNAMED_TABLE,
)
from .shared import CMYKValue, DomainData, PaletteItem, RGBValue, ValueTables

GMT_CPT_KEYWORD = (
    "#",
    "COLOR_MODEL",
    "B",
    "F",
    "N",
    CM_CMYK,
    CM_GRAY,
    CM_HSV,
    CM_RGB,
)
GMT_CPT_COMMENT = GMT_CPT_KEYWORD[0]
GMT_CPT_COLOR_MODEL = {CM_CMYK, CM_GRAY, CM_HSV, CM_RGB}


class cpt_utility(clr_utility):

    @classmethod
    def parse_cpt_table(
        cls, lines: list[str]
    ) -> tuple[PaletteItem, str, DomainData]:
        j: list[float] = []

        r: list[float] = []
        g: list[float] = []
        b: list[float] = []
        k: list[float] = []

        clr_values = r, g, b, k

        cl_nan = float("nan")
        bg: CMYKValue = (cl_nan, cl_nan, cl_nan, cl_nan)
        fg: CMYKValue = (cl_nan, cl_nan, cl_nan, cl_nan)
        nn: CMYKValue = 1.0, 0.0, 1.0, cl_nan

        color_model = CM_RGB

        for line in lines:
            # Split line into list of strings of keywords or values
            ls = line.split()

            # Check for alternative color model
            if line[0] == GMT_CPT_COMMENT and ls[-1] in GMT_CPT_COLOR_MODEL:
                color_model = ls[-1]

            # Extract stock colors
            bg, fg, nn = cls._extract_stock_color(color_model, bg, fg, nn, ls)

            # Ignore header lines
            if ls[0] in GMT_CPT_KEYWORD:
                continue

            cls._extract_color_range(j, clr_values, color_model, ls)

        bg, fg = cls._finalize_stock_colors(clr_values, bg, fg, color_model)

        # The `r`, `g`, and `b` lists are modified in place and contain
        # the normalized color component intensity values corresponding
        # to the blue, green, and red color components, respectively;
        # the `n` list is left unchanged since it is used only for CMYK
        # to RGB colorspace conversion.
        r, g, b = cls._process_cpt_colors(color_model, r, g, b, k)

        extent = j[0], j[-1]

        entries = cls._make_color_entries(j, b, g, r)
        stock = cls._process_cpt_stock(color_model, bg, fg, nn)

        return (entries, stock), UNNAMED_TABLE, extent

    @staticmethod
    def _cmyk_to_rgb(c: float, m: float, y: float, k: float) -> RGBValue:
        b = 1.0 - k / CMYK_MAX
        return (
            (1.0 - c / CMYK_MAX) * b,
            (1.0 - m / CMYK_MAX) * b,
            (1.0 - y / CMYK_MAX) * b,
        )

    @classmethod
    def _extract_color_range(
        cls,
        j: list[float],
        clr_values: ValueTables,
        color_model: str,
        ls: list[str],
    ) -> None:
        r, g, b, k = clr_values

        if color_model == CM_GRAY:
            j.extend((float(ls[0]), float(ls[2])))
            r.extend((float(ls[1]), float(ls[3])))

        elif color_model == CM_CMYK:
            j.extend((float(ls[0]), float(ls[5])))
            r.extend((float(ls[1]), float(ls[6])))
            g.extend((float(ls[2]), float(ls[7])))
            b.extend((float(ls[3]), float(ls[8])))
            k.extend((float(ls[4]), float(ls[9])))

        else:
            j.extend((float(ls[0]), float(ls[4])))
            r.extend((float(ls[1]), float(ls[5])))
            g.extend((float(ls[2]), float(ls[6])))
            b.extend((float(ls[3]), float(ls[7])))

    @classmethod
    def _extract_stock_color(
        cls,
        color_model: str,
        bg: CMYKValue,
        fg: CMYKValue,
        nn: CMYKValue,
        ls: list[str],
    ) -> tuple[CMYKValue, CMYKValue, CMYKValue]:
        if ls[0] == GMT_CPT_KEYWORD[2]:
            bg = cls._get_stock_color(color_model, ls)
        elif ls[0] == GMT_CPT_KEYWORD[3]:
            fg = cls._get_stock_color(color_model, ls)
        elif ls[0] == GMT_CPT_KEYWORD[4]:
            nn = cls._get_stock_color(color_model, ls)
        return bg, fg, nn

    @classmethod
    def _finalize_stock_colors(
        cls,
        clr_values: ValueTables,
        bg: CMYKValue,
        fg: CMYKValue,
        color_model: str,
    ) -> tuple[CMYKValue, CMYKValue]:
        r, g, b, k = clr_values

        if color_model == CM_GRAY:
            bg = (r[0], 0.0, 0.0, 0.0) if math.isnan(bg[0]) else bg
            fg = (r[-1], 0.0, 0.0, 0.0) if math.isnan(fg[0]) else fg

        elif color_model == CM_CMYK:
            bg = (r[0], g[0], b[0], k[0]) if math.isnan(bg[0]) else bg
            fg = (r[-1], g[-1], b[-1], k[-1]) if math.isnan(fg[0]) else fg

        else:
            bg = (r[0], g[0], b[0], 0.0) if math.isnan(bg[0]) else bg
            fg = (r[-1], g[-1], b[-1], 0.0) if math.isnan(fg[0]) else fg

        return bg, fg

    @staticmethod
    def _get_stock_color(color_model: str, ls: list[str]) -> CMYKValue:
        r, g, b, k = float(ls[1]), 0.0, 0.0, 0.0
        if color_model in {CM_RGB, CM_CMYK}:
            g = float(ls[2])
            b = float(ls[3])
            if color_model == CM_CMYK:
                k = float(ls[4])
        return r, g, b, k

    @staticmethod
    def _hsv_to_rgb(h: float, s: float, v: float) -> RGBValue:
        return colorsys.hsv_to_rgb(h / HUE_MAX, s / HSV_MAX, v / HSV_MAX)

    @staticmethod
    def _normalize_grayscale(x: float) -> float:
        return x / BRG_MAX

    @classmethod
    def _process_cpt_colors(
        cls,
        color_model: str,
        r: list[float],
        g: list[float],
        b: list[float],
        n: list[float],
    ) -> ValueTables:
        # Normalize color values
        if color_model == CM_RGB:
            r = list(map(cls._normalize_color, r))
            g = list(map(cls._normalize_color, g))
            b = list(map(cls._normalize_color, b))

        # Convert color model if necessary
        elif color_model == CM_HSV:
            for i, (h, s, v) in enumerate(zip(r, g, b)):
                r[i], g[i], b[i] = cls._hsv_to_rgb(h, s, v)

        elif color_model == CM_CMYK:
            for i, (c, m, y, k) in enumerate(zip(r, g, b, n)):
                r[i], g[i], b[i] = cls._cmyk_to_rgb(c, m, y, k)

        elif color_model == CM_GRAY:
            r = list(map(cls._normalize_grayscale, r))
            g, b = r, r

        else:
            raise ValueError("Invalid color model")

        return r, g, b

    @classmethod
    def _process_cpt_stock(
        cls,
        color_model: str,
        bg: CMYKValue,
        fg: CMYKValue,
        nn: CMYKValue,
    ) -> list[RGBValue]:
        packed = (bg, fg, nn)
        u, v, w, y = zip(*packed)
        r, g, b, k = list(u), list(v), list(w), list(y)

        r, g, b = cls._process_cpt_colors(color_model, r, g, b, k)

        return list(zip(r, g, b))



"""
Export locators for GridSat imagery products.

Classes:
    - GridSatProductLocatorB1: Locator for B1 products.
    - GridSatProductLocatorGC: Locator for GC products.
"""

from .locator_b1 import GridSatProductLocatorB1
from .locator_gc import GridSatProductLocatorGC
from .netcdf_geodetic import GSLatLonGrid
from .netcdf_image import GSImage
from .netcdf_metadata import GSDatasetMetadata
from .netcdf_time import GSCoverageTime, GSTimeGrid
from .preview import GSPlot, GSPlotParameter, read_gridsat_dataset

__all__ = [
    "GridSatProductLocatorB1",
    "GridSatProductLocatorGC",
    "GSPlot",
    "GSImage",
    "GSLatLonGrid",
    "GSTimeGrid",
    "GSCoverageTime",
    "GSPlotParameter",
    "GSDatasetMetadata",
    "read_gridsat_dataset",
]



"""
GOES precomputed latitude and longitude grids data extraction.

This module provides classes to extract and represent latitude and
longitude grids information from GOES satellite netCDF data files.

Classes
-------
GOESGeodeticGrid
    Hold GOES satellite precomputed latitude and longitude data.
GOESLatLonGridData
    Represent GOES satellite precomputed latitude or longitude data.
GOESLatLonGridMetadata
    Hold GOES satellite latitude and longitude metadata.
GOESLatLonGridMetadataType
    Represent a GOES satellite latitude or longitude metadata.
GOESLatLonGridInfo
    Hold GOES geodetic grid dataset metadata information.
"""

from re import match
from typing import Any, cast

from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from numpy import errstate, float32, nan
from numpy.ma import masked_invalid

from ..geodesy import (
    GeostationaryParameters,
    calculate_pixel_edges,
    geos_to_latlon_grid_cartopy,
    geos_to_latlon_grid_goesdl,
    geos_to_latlon_grid_pyproj,
)
from ..netcdf import DatasetView, HasStrHelp, dimension, variable
from ..utils.array import ArrayBool, ArrayFloat32, ArrayFloat64, MaskedFloat32
from .netcdf_projection import GOESProjection


class GOESLatLonGridData:
    """
    Represent GOES satellite latitude or longitude grid data.

    Attributes
    ----------
    data : ArrayFloat32
        The latitude or longitude grid data.
    mask : NDArray[bool]
        The array containing the mask values indicating invalid data
        points.
    fill_value : uint16
        The fill value used for missing or invalid data points.
    """

    def __init__(self, array: MaskedFloat32) -> None:
        """
        Initialize a GOESLatLonGridData object.

        Parameters
        ----------
        array : MaskedFloat32
            The masked array containing the latitude or longitude grid
            data.
        """
        self.data = array.data
        self.mask = array.mask
        self.fill_value = array.fill_value

    data: ArrayFloat32
    mask: ArrayBool
    fill_value: float32


def _latlon_data(
    name: str, record: Dataset, step: tuple[int, int] | None
) -> GOESLatLonGridData:
    latlon = variable(name)

    def _subsample(x: Any) -> Any:
        return x[:] if step is None else x[:: step[0], :: step[1]]

    class _GOESLatLonGridData(DatasetView):
        data: ArrayFloat32 = latlon.data(filter=_subsample)
        mask: ArrayBool = latlon.mask(filter=_subsample)
        fill_value: float32 = latlon.fill_value()

    _GOESLatLonGridData.__module__ = GOESLatLonGridData.__module__
    _GOESLatLonGridData.__name__ = GOESLatLonGridData.__name__
    _GOESLatLonGridData.__qualname__ = GOESLatLonGridData.__qualname__

    data = _GOESLatLonGridData(record)

    return cast(GOESLatLonGridData, data)


class GOESLatLonGrid(HasStrHelp):
    """
    Hold GOES satellite precomputed latitude and longitude data.

    Notes
    -----
    For information on GOES Imager Projection and GOES orbit geometry,
    see [1]_ and Section 4.2. of [2]_ For information on getting the
    precomputed latitude and longitude grid dataset, see [1]_.

    Attributes
    ----------
    latitude : GOESLatLonGridData
        The latitude grid data.
    longitude : GOESLatLonGridData
        The longitude grid data.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    latitude: GOESLatLonGridData
    longitude: GOESLatLonGridData

    def __init__(self, record: Dataset) -> None:
        """
        Initialize a GOESGrid object from a precomputed netCDF dataset.

        Parameters
        ----------
        record : Dataset
            The netCDF dataset containing the precomputed latitude
            and longitude grid data.
        """
        self.latitude = _latlon_data("latitude", record, None)
        self.longitude = _latlon_data("longitude", record, None)


class GOESLatLonGridMetadataType:
    """
    Represent a GOES satellite latitude or longitude metadata.

    Attributes
    ----------
    long_name : str
        A descriptive name for the variable.
    valid_range : ArrayFloat64
        The valid range of values for the variable.
    units : str
        The units of measurement for the variable.
    comment : str
        Additional comments or information about the variable.
    """

    long_name: str
    valid_range: ArrayFloat64
    units: str
    comment: str


def _latlon_metadata(name: str, record: Dataset) -> GOESLatLonGridMetadataType:
    latlon = variable(name)

    class _GOESLatLonGridMetadata(DatasetView):
        long_name: str = latlon.attribute()
        valid_range: ArrayFloat64 = latlon.attribute()
        units: str = latlon.attribute()
        comment: str = latlon.attribute()

    _GOESLatLonGridMetadata.__module__ = GOESLatLonGridMetadataType.__module__
    _GOESLatLonGridMetadata.__name__ = GOESLatLonGridMetadataType.__name__
    _GOESLatLonGridMetadata.__qualname__ = (
        GOESLatLonGridMetadataType.__qualname__
    )

    metadata = _GOESLatLonGridMetadata(record)

    return cast(GOESLatLonGridMetadataType, metadata)


class GOESLatLonGridMetadata(HasStrHelp):
    """
    Hold a GOES satellite latitude and longitude metadata.

    Attributes
    ----------
    latitude : GOESLatLonGridMetadataType
        The latitude metadata.
    longitude : GOESLatLonGridMetadataType
        The longitude metadata.
    """

    latitude: GOESLatLonGridMetadataType
    longitude: GOESLatLonGridMetadataType

    def __init__(self, record: Dataset) -> None:
        self.latitude = _latlon_metadata("latitude", record)
        self.longitude = _latlon_metadata("longitude", record)


class GOESLatLonGridInfo(DatasetView):
    """
    Hold GOES geodetic grid dataset metadata information.

    Attributes
    ----------
    title : str
        The title of the dataset.
    comment : str
        Comments or information about the dataset.
    created : str
        The date the dataset was created.
    rows : int
        The number of rows in the the grid.
    columns : int
        The number of columns in the the grid.
    """

    title: str
    comment: str
    created: str
    rows: int = dimension.size()
    columns: int = dimension.size()


FILL_VALUE = -999.99


class GOESGeodeticGrid(HasStrHelp):
    """
    Hold GOES satellite geodetic grids data.

    Notes
    -----
    For information on GOES Imager Projection and GOES orbit geometry,
    see [1]_ and Section 4.2. of [2]_. GOES-R use the Geodetic Reference
    System 1980 (GRS80).

    Attributes
    ----------
    latitude : GOESLatLonGridData
        The GRS80 latitude grid data.
    longitude : GOESLatLonGridData
        The GRS80 longitude grid data.

    References
    ----------
    .. [1] STAR Atmospheric Composition Product Training, "GOES Imager
        Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
        https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
    .. [2] GOES-R, " GOES-R Series Product Definition and User’s Guide
        (PUG), Volume 5: Level 2+ Products", Version 2.4,
        NASA/NOAA/NESDIS, 2022.
        https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
    """

    latitude: GOESLatLonGridData
    longitude: GOESLatLonGridData

    def __init__(
        self, latitude: GOESLatLonGridData, logitude: GOESLatLonGridData
    ) -> None:
        """
        Initialize a GOESGrid object from a precomputed netCDF dataset.

        Parameters
        ----------
        latitude : GOESLatLonGridData
            The GRS80 latitude grid data.
        longitude : GOESLatLonGridData
            The GRS80 longitude grid data
        """
        self.latitude = latitude
        self.longitude = logitude

    @classmethod
    def calculate(
        cls,
        record: Dataset,
        algorithm: str = "goesdr",
        step: int | tuple[int, int] = 1,
    ) -> "GOESGeodeticGrid":
        """
        Compute the geodetic grids and initialize a GOESGrid object.

        Calculate the latitude and longitude grid arrays on the fly from
        the GOES Imager Projection information in ABI Level 2 files.

        Notes
        -----
        For information on GOES Imager Projection and GOES orbit
        geometry, see [1]_ and Section 4.2. of [3]_. For a Python
        demonstration on calculating latitude and longitude from GOES
        Imager Projection information, see [2]_. The code snippet in
        this class is an optimized version based on the code from [2]_.

        Parameters
        ----------
        record : Dataset
            The netCDF dataset containing ABI Level 2 data.
        algorithm : str, optional
            The algorithm to use to calculate the latitude and longitude
            grid data. Choose 'cartopy', 'goesdr', or 'pyproj'.
            'cartopy' or 'pyproj' requires the respective Python package
            to be installed.  'goesdr' is the default algorithm and does
            not require any additional packages.
        step : int or tuple[int, int], optional
            The step size to subsample the latitude and longitude grid
            data. If an integer is provided, the step size is the same
            for both rows and columns. If a tuple is provided, the first
            value is the step size for grid rows and the second value is
            the step size for grid columns. The default is 1.

        Returns
        -------
        GOESLatLonGrid
            The latitude and longitude grid data.

        References
        ----------
        .. [1] STAR Atmospheric Composition Product Training, "GOES
            Imager Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
            https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
        .. [2] Aerosols and Atmospheric Composition Science Team,
            "Python Short Demo: Calculate Latitude and Longitude from
            GOES Imager Projection (ABI Fixed Grid) Information",
            NOAA/NESDIS/STAR, 2024.
            https://www.star.nesdis.noaa.gov/atmospheric-composition-training/python_abi_lat_lon.php
        .. [3] GOES-R, " GOES-R Series Product Definition and User’s
            Guide (PUG), Volume 5: Level 2+ Products", Version 2.4,
            NASA/NOAA/NESDIS, 2022.
            https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
        """
        algorithm, corners = cls._parse_algorithm(algorithm)

        step = cls._parse_step(step)

        abi_lat, abi_lon = cls._initialize_calculated(
            record, algorithm, corners, step
        )

        return GOESGeodeticGrid(abi_lat, abi_lon)

    @staticmethod
    def from_file(
        record: Dataset, step: int | tuple[int, int] | None = None
    ) -> "GOESGeodeticGrid":
        """
        Initialize a GOESGrid object from a precomputed netCDF dataset.

        Notes
        -----
        For information on GOES Imager Projection and GOES orbit
        geometry, see [1]_ and Section 4.2. of [2]_. For information on
        getting the precomputed latitude and longitude grids dataset,
        see [1]_.

        Parameters
        ----------
        record : Dataset
            The netCDF dataset containing precomputed latitude and
            longitude grid data.
        step : int or tuple[int, int], optional
            The step size to subsample the latitude and longitude grid
            data. If an integer is provided, the step size is the same
            for both rows and columns. If a tuple is provided, the first
            value is the step size for grid rows and the second value is
            the step size for grid columns. The default is None.

        Returns
        -------
        GOESLatLonGrid
            The latitude and longitude grid data.

        References
        ----------
        .. [1] STAR Atmospheric Composition Product Training, "GOES
            Imager Projection (ABI Fixed Grid)", NOAA/NESDIS/STAR, 2024.
            https://www.star.nesdis.noaa.gov/atmospheric-composition-training/satellite_data_goes_imager_projection.php.
        .. [2] GOES-R, " GOES-R Series Product Definition and User’s
            Guide (PUG), Volume 5: Level 2+ Products", Version 2.4,
            NASA/NOAA/NESDIS, 2022.
            https://www.ospo.noaa.gov/Organization/Documents/PUG/GS%20Series%20416-R-PUG-L2%20Plus-0349%20Vol%205%20v2.4.pdf
        """
        if isinstance(step, int):
            step = (step, step)

        abi_lat = _latlon_data("latitude", record, step)
        abi_lon = _latlon_data("longitude", record, step)

        abi_lat.data[abi_lat.mask] = nan
        abi_lon.data[abi_lon.mask] = nan

        return GOESGeodeticGrid(abi_lat, abi_lon)

    @classmethod
    def _initialize_calculated(
        cls,
        record: Dataset,
        algorithm: str,
        corners: bool,
        step: tuple[int, int],
    ) -> tuple[GOESLatLonGridData, GOESLatLonGridData]:
        projection_info = GOESProjection(record)

        if corners:
            x_r = calculate_pixel_edges(projection_info.x)
            y_r = calculate_pixel_edges(projection_info.y)
        else:
            x_r = projection_info.x
            y_r = projection_info.y

        if step[0] > 1:
            y_r = y_r[:: step[0]]

        if step[1] > 1:
            x_r = x_r[:: step[1]]

        orbit_parameters = (
            projection_info.longitude_of_projection_origin,
            projection_info.perspective_point_height,
            projection_info.sweep_angle_axis,
        )
        globe_parameters = (
            projection_info.semi_major_axis,
            projection_info.semi_minor_axis,
            projection_info.inverse_flattening,
        )
        xy_grid = (x_r, y_r)

        parameters = GeostationaryParameters(
            orbit_parameters, globe_parameters, xy_grid
        )

        lat, lon = cls._calculate_latlon_grid(parameters, algorithm)

        abi_lat = GOESLatLonGridData(lat)
        abi_lon = GOESLatLonGridData(lon)

        return abi_lat, abi_lon

    @staticmethod
    def _calculate_latlon_grid(
        parameters: GeostationaryParameters,
        algorithm: str,
    ) -> tuple[MaskedFloat32, MaskedFloat32]:
        # Ignore numpy errors for `sqrt` of negative number; reference
        # [2] states that this "occurs for GOES-16 ABI CONUS sector
        # data", however I found it also occurs for Full Disk sector.
        with errstate(invalid="ignore"):
            if algorithm == "goesdr":
                lat, lon = geos_to_latlon_grid_goesdl(parameters)
            elif algorithm == "pyproj":
                lat, lon = geos_to_latlon_grid_pyproj(parameters)
            elif algorithm == "cartopy":
                lat, lon = geos_to_latlon_grid_cartopy(parameters)
            else:
                raise ValueError(
                    f"Invalid algorithm '{algorithm}'. Expected pattern: "
                    "'<algorithm>' or '<algorithm>[<option>]'. "
                    "Choose 'precomputed', 'noaa', 'opti' (default), "
                    "'pyproj', or 'cartopy' for '<algorithm>'. Choose "
                    "'center' (default) or 'corner' for '<option>'. "
                    "'precomputed' cannot be used with options."
                )

        latitude: MaskedFloat32 = masked_invalid(lat)  # type: ignore
        longitude: MaskedFloat32 = masked_invalid(lon)  # type: ignore

        latitude.fill_value = FILL_VALUE
        longitude.fill_value = FILL_VALUE

        return latitude, longitude

    @staticmethod
    def _parse_algorithm(algorithm: str) -> tuple[str, bool]:
        pattern = r"^(\w+)(?:\[(\w+)\])?$"
        match_ = match(pattern, algorithm)

        if not match_:
            raise ValueError(
                f"Invalid algorithm '{algorithm}'. Expected pattern: "
                "'<algorithm>' or '<algorithm>[<option>]'. "
                "Choose 'cartopy', 'goesdr' (default), or 'pyproj' for "
                "'<algorithm>'. Choose 'center' (default) or 'corner' "
                "for '<option>'."
            )

        name = match_[1]

        if name not in {"cartopy", "goesdr", "pyproj"}:
            raise ValueError(
                f"Invalid algorithm '{name}'. Choose 'cartopy', 'goesdr', "
                "or 'pyproj'."
            )

        option = match_[2]

        if option not in {None, "center", "corner"}:
            raise ValueError(
                f"Invalid option '{option}'. Choose 'center' or 'corner'."
            )

        corner: bool = option == "corner"

        return name, corner

    @staticmethod
    def _parse_step(step: int | tuple[int, int]) -> tuple[int, int]:
        if isinstance(step, int):
            step = (step, step)

        if step[0] <= 0 or step[1] <= 0:
            raise ValueError(f"Step size must be greater than 0, got {step}.")

        return step



from numpy import float64


class AeqdProjParameters:
    """
    A class to store Azimuthal Equidistant projection parameters.

    Attributes
    ----------
    longitude_of_projection_origin : float64
        The longitude of the projection origin in East degrees.
    latitude_of_projection_origin : float64
        The latitude of the projection origin in North degrees.
    extent_size : tuple[float64, float64, str]
        The size of the extent in the specified unit of measurement:
        - The width of the extent in the specified unit of measurement.
        - The height of the extent in the specified unit of measurement.
        - The unit of measurement; e.g., 'm' for meters, 'km' for
          kilometers, 'arcsec' for arc-seconds, 'arcmin' for arc-
          minutes, or 'deg' for degrees.
    """

    # Information about the projection
    longitude_of_projection_origin: float64
    latitude_of_projection_origin: float64
    extent_size: tuple[float64, float64, str]

    def __init__(
        self,
        longitude_of_projection_origin: float64,
        latitude_of_projection_origin: float64,
        extent_size: tuple[float64, float64, str],
    ) -> None:
        """
        Initialize the class.

        Parameters
        ----------
        longitude_of_projection_origin : float64
            The longitude of the projection origin in East degrees.
        latitude_of_projection_origin : float64
            The latitude of the projection origin in North degrees.
        extent_size : tuple[float64, float64, str]
            The size of the extent in the specified unit of measurement.
            The first element is the width, the second is the height,
            and the third is the unit of measurement; e.g. 'm' for
            meters, 'km' for kilometers, 'arcsec' for arc seconds,
            'arcmin' for arc minutes, or 'deg' for degrees.
        """
        self.longitude_of_projection_origin = longitude_of_projection_origin
        self.latitude_of_projection_origin = latitude_of_projection_origin
        self.extent_size = extent_size

        units = extent_size[2]
        if units not in {"m", "km", "arcsec", "arcmin", "deg"}:
            raise ValueError(
                f"Invalid unit of measurement: {units}. "
                "Valid options are 'm', 'km', 'arcsec', 'arcmin', or 'deg'."
            )



"""
Define the class for the downloader object.

The downloader object is responsible for downloading files from a
specified location using a datasource and product locator.

Classes
-------
Downloader
    Represent a downloader object.
"""

from pathlib import Path

from ..dataset import ProductLocator
from ..datasource import Datasource
from ..utils import FileRepository
from .constants import ISO_TIMESTAMP_FORMAT, TIME_TOLERANCE_DEFAULT
from .downloader_base import DownloaderBase


class Downloader(DownloaderBase):
    """
    Represent a downloader object.

    The downloader object is responsible for downloading files from a
    datasource and store it in a local repository. It receives a
    `Datasource` object, a `ProductLocator` object, and a
    `FileRepository` object. Optional parameters are a date format
    specification, a time tolerance value and a flag to show the
    download progress. It provides methods to download files from the
    datasource that match the timestamps between a start and end time.

    The downloader object is agnostic to the datasource, dataset, and
    dataset's product implementations. It only requires that the
    datasource and product locator objects implement the methods
    specified in the `goesdl.datasource.Datasource`,
    `goesdl.dataset.ProductLocator` and `goesdl.dataset.Product`
    interfaces.
    """

    def __init__(
        self,
        datasource: Datasource,
        locator: ProductLocator,
        repository: str | Path | FileRepository | None = None,
        date_format: str = ISO_TIMESTAMP_FORMAT,
        time_tolerance: int = TIME_TOLERANCE_DEFAULT,
        show_progress: bool = True,
    ) -> None:
        """
        Initialize the downloader object.

        Parameters
        ----------
        datasource : Datasource
            The datasource object to use for downloading files.
        locator : ProductLocator
            The product locator object to use for locating files.
        repository : str | Path | FileRepository | None
            The repository to store the downloaded files. If a string or
            Path object is provided, a FileRepository object is created
            using the provided path. If None, a FileRepository object is
            created using the current working directory. Default is
            None.
        date_format : str
            The date format to use for parsing timestamps. Default is
            'iso'.
        time_tolerance : int
            The time tolerance value to use for matching timestamps.
            Default is 5.
        show_progress : bool
            A flag to show the download progress. Default is True.
        """
        if not isinstance(repository, FileRepository):
            repository = FileRepository(repository)

        super().__init__(
            datasource=datasource,
            locator=locator,
            repository=repository,
            date_format=date_format,
            time_tolerance=time_tolerance,
            show_progress=show_progress,
        )



"""
Export class providing the base functionalities for product locators.

Classes:
    - ProductLocator: Abstract a product locator for satellite imagery
      datasets.
    - ProductLocatorGG: Abstract a GridSat or GOES-R series dataset
      product locator.
"""

from .locator import ProductLocator
from .locator_gg import ProductLocatorGG

__all__ = ["ProductLocator", "ProductLocatorGG"]



"""
Contain test functions for the GOES-DL project.

These functions are responsible for testing the functionality of the
downloader objects with different product locators and datasources.

Attributes
----------
DATE_FORMAT : str
    The date format specification used for parsing dates.
REPO_GRISAT_B1 : str
    The repository path for GridSat-B1 data.
REPO_GRISAT_GOES : str
    The repository path for GridSat-GOES/CONUS data.
REPO_GOES : str
    The repository path for GOES-16 data.

Functions
---------
test(dl: Downloader, start: str, end: str = "")
    Test the downloader object by downloading files within a date range.
test_gridsat_aws()
    Test the downloader object with GridSat-B1 data and AWS datasource.
test_gridsat_http()
    Test the downloader object with GridSat-B1 data and HTTP datasource.
test_gridsat_goes()
    Test the downloader object with GridSat-GOES/CONUS data and HTTP
    datasource.
test_goes()
    Test the downloader object with GOES-16 data and AWS datasource.
"""

from goesdl.datasource import DatasourceAWS, DatasourceLocal, DatasourceNCEI
from goesdl.downloader import Downloader
from goesdl.goesr import GOESProductLocatorABIPP as ProductLocatorGOES
from goesdl.gridsat import GridSatProductLocatorB1 as ProductLocatorB1
from goesdl.gridsat import GridSatProductLocatorGC as ProductLocatorGC

DATE_FORMAT = "%Y-%m-%dT%H:%M%z"


REPO_GRISAT_B1 = "../temp/gridsat/B1"
REPO_GRISAT_GOES = "../temp/gridsat/goes-conus"
REPO_GOES = "../temp/GOES-16"


def test(dl: Downloader, start: str, end: str = "") -> list[str]:
    """
    Test the downloader object by downloading files within a date range.

    Parameters
    ----------
    dl : Downloader
        The downloader object to test.
    start : str
        The start date of the download range.
    end : str, optional
        The end date of the download range, if any. The default is "".

    Returns
    -------
    list[str]
        The list of downloaded files.
    """
    if end:
        print(f"Downloading data from {start} to {end}")
    else:
        print(f"Downloading data from {start}")

    return dl.download_files(start=start, end=end)


def test_gridsat_aws() -> list[str]:
    """
    Test the downloader object with GridSat-B1 data and AWS datasource.

    Returns
    -------
    list[str]
        The list of downloaded files.
    """
    pd = ProductLocatorB1()
    ds = DatasourceAWS(pd.get_base_url("AWS"))

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=REPO_GRISAT_B1,
        date_format=DATE_FORMAT,
    )

    return test(dl, "1984-08-23T00:00Z")


def test_gridsat_http() -> tuple[list[str], list[str]]:
    """
    Test the downloader object with GridSat-B1 data and HTTP datasource.

    Returns
    -------
    tuple[list[str], list[str]]
        A tuple of lists of downloaded files.
    """
    pd = ProductLocatorB1()
    ds = DatasourceNCEI(pd)

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=REPO_GRISAT_B1,
        date_format=DATE_FORMAT,
    )

    files1 = test(dl, "1984-08-23T00:00Z")
    files2 = test(dl, "1982-08-23T00:00Z")

    return files1, files2


def test_gridsat_goes() -> list[str]:
    """
    Test the downloader with GridSat-GOES/CONUS and HTTP datasource.

    Returns
    -------
    list[str]
        The list of downloaded files.
    """
    pd = ProductLocatorGC("F", ["G11", "G12"])
    ds = DatasourceNCEI(pd)

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=REPO_GRISAT_GOES,
        date_format=DATE_FORMAT,
    )

    return test(dl, "2007-09-02T18:00Z")


def test_goes1() -> list[str]:
    """
    Test the downloader object with GOES-16 data and AWS datasource.

    Returns
    -------
    list[str]
        The list of downloaded files.
    """
    pd = ProductLocatorGOES("CMIP", "F", "C13", "G16")

    # GOES-16 data is updated every 10 minutes. If you are downloading
    # old data, you may leave the refresh rate as default.

    ds = DatasourceAWS(pd, 10 * 60)

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=REPO_GOES,
        date_format=DATE_FORMAT,
    )

    return test(dl, "2024-08-23T00:00+0000")


def test_goes2() -> tuple[list[str], list[str]]:
    """
    Test the downloader object with GOES-16 data and AWS datasource.

    Returns
    -------
    tuple[list[str], list[str]]
        A tuple of lists of downloaded files.
    """
    pd = ProductLocatorGOES("CMIP", "F", "C13", "G16")

    # GOES-16 data is updated every 10 minutes. If you are downloading
    # old data, you may leave the refresh rate as default.

    ds = DatasourceAWS(pd, 10 * 60)

    repo_goes_l = "../../TFG_Tools/repository/20201114T20"

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=repo_goes_l,
        date_format=DATE_FORMAT,
    )

    files1 = test(dl, "2020-11-14T20:00Z")
    files2 = test(dl, "2020-11-14T20:00Z", "2020-11-15T19:00Z")

    return files1, files2


def test_goes3() -> tuple[list[str], list[str]]:
    """
    Test the downloader object with GOES-16 data and local datasource.

    Returns
    -------
    tuple[list[str], list[str]]
        A tuple of lists of downloaded files.
    """
    pd = ProductLocatorGOES("CMIP", "F", "C13", "G16")

    repo_goes_s = "../../TFG_Tools/repository/20201114T20"

    ds = DatasourceLocal(repo_goes_s, 0)

    repo_goes_d = "../../TFG_Tools/repository/20201114T21"

    dl = Downloader(
        datasource=ds,
        locator=pd,
        repository=repo_goes_d,
        date_format=DATE_FORMAT,
    )

    files1 = test(dl, "2020-11-14T20:00Z")
    files2 = test(dl, "2020-11-14T20:00Z", "2020-11-15T19:00Z")

    return files1, files2


def main() -> None:
    """Run all test functions."""
    files2 = test_goes3()
    print(files2)

    files2 = test_goes2()
    print(files2)

    files = test_goes1()
    print(files)

    files = test_gridsat_goes()
    print(files)

    files = test_gridsat_aws()
    print(files)

    files2 = test_gridsat_http()
    print(files2)


if __name__ == "__main__":
    main()



from collections.abc import Generator
from math import degrees, floor
from typing import Literal, cast

from numpy import arange, float32

from ..utils.array import ArrayFloat32

CoordDomain = tuple[float, float, float, float]
CoordLimits = tuple[float, float]
ExtentType = tuple[float, float]
CenterType = tuple[float, float]
StepType = tuple[float, float]

DEFAULT_STEP = (2.0, 2.0)


class RectangularRegion:

    domain: CoordDomain

    xticks: ArrayFloat32
    yticks: ArrayFloat32

    def __init__(self, domain: CoordDomain) -> None:
        self.domain = domain

        self.xticks, self.yticks = self._create_grid_ticks(
            domain, DEFAULT_STEP
        )

    def set_ticks(
        self, domain: CoordDomain | None = None, step: StepType = DEFAULT_STEP
    ) -> None:
        domain = self.domain if domain is None else domain
        self.xticks, self.yticks = self._create_grid_ticks(domain, step)

    @classmethod
    def from_central_point(
        cls,
        extent: ExtentType,
        center_deg: CenterType,
        units: Literal["arcsec", "arcmin", "deg", "rad"] = "deg",
    ) -> "RectangularRegion":
        extent_deg: ExtentType | Generator[float]
        if units == "arcsec":
            extent_deg = (value / 3600.0 for value in extent)
        elif units == "arcmin":
            extent_deg = (value / 60.0 for value in extent)
        elif units == "rad":
            extent_deg = (degrees(value) for value in extent)
        else:
            extent_deg = extent

        width_deg, height_deg = extent_deg
        lon_cen, lat_cen = center_deg

        x_half = 0.5 * width_deg
        y_half = 0.5 * height_deg

        lons_deg = [lon_cen - x_half, lon_cen + x_half]
        lats_deg = [lat_cen - y_half, lat_cen + y_half]

        domain = tuple(lons_deg + lats_deg)

        return cls(cast(CoordDomain, domain))

    @staticmethod
    def _create_grid_ticks(
        domain: CoordDomain, step: StepType
    ) -> tuple[ArrayFloat32, ArrayFloat32]:
        dx, dy = step
        lon_min = dx * floor(domain[0] / dx)
        lon_max = dx * floor(domain[1] / dx) + dx
        xticks = arange(lon_min, lon_max, dx).astype(float32)

        lat_min = dy * floor(domain[2] / dy)
        lat_max = dy * floor(domain[3] / dy) + dy
        yticks = arange(lat_min, lat_max, dy).astype(float32)

        return cast(ArrayFloat32, xticks), cast(ArrayFloat32, yticks)

    @property
    def extent(self) -> CoordDomain:
        return self.domain

    @property
    def lon_bounds(self) -> CoordLimits:
        return self.domain[:2]

    @property
    def lat_bounds(self) -> CoordLimits:
        return self.domain[2:]



"""
Provide a utility class for URL manipulations.

Classes:
    url: Utility class for URL manipulations.
"""

from urllib.parse import ParseResult, urljoin, urlparse


class URL:
    """
    Utility class for URL manipulations.

    Methods
    -------
    abspath(path: str)
        Returns the absolute path of the given URL.
    join(base_url: str, *parts: str)
        Joins one or more URL parts to the base URL.
    parse(path: str)
        Parses the given URL and returns a ParseResult object.
    """

    @staticmethod
    def abspath(path: str) -> str:
        """
        Return the absolute path of the given URL.

        Parameters
        ----------
        path : str
            The URL to be converted to an absolute path.

        Returns
        -------
        str
            The absolute path of the given URL.
        """
        parsed_url: ParseResult = urlparse(path)
        abs_path = urljoin(parsed_url.netloc, parsed_url.path)
        parsed_url = parsed_url._replace(path=abs_path)
        return parsed_url.geturl()

    @staticmethod
    def join(base_url: str, *parts: str) -> str:
        """
        Join one or more URL parts to the base URL.

        Parameters
        ----------
        base_url : str
            The base URL to which parts will be joined.
        *parts : str
            One or more URL parts to be joined to the base URL.

        Returns
        -------
        str
            The resulting URL after joining the parts to the base URL.
        """
        dest_url: str = base_url
        for part in parts:
            dest_url = urljoin(dest_url, part)
        return dest_url

    @staticmethod
    def parse(path: str) -> ParseResult:
        """
        Parse the given URL and returns a ParseResult object.

        Parameters
        ----------
        path : str
            The URL to be parsed.

        Returns
        -------
        ParseResult
            The parsed URL as a ParseResult object.
        """
        return urlparse(path)



"""
Provide utility functions for generating type hints.

Generate type hints for various collections and data types. The type
hints are generated as strings that can be used for type annotations in
Python code. Include special handling for NumPy arrays and homogeneous
collections.

Functions
---------
get_value_typehint(value: Any) -> str
    Generate a type hint for a given value.
"""

from collections.abc import Iterable, Sequence
from typing import Any

from numpy import dtype, ndarray


def _get_multitype_typehint(multitype: Iterable[Any], union: str) -> str:
    # Generate a type hint for a collection with multiple types.
    type_hints: set[str] = set()
    for element in multitype:
        type_hint = get_typehint(element)
        type_hints.add(type_hint)
    return union.join(type_hints)


def _get_iterable_typehint(iterable: Iterable[Any]) -> str:
    # Generate a type hint for an iterable collection.
    return _get_multitype_typehint(iterable, " | ") or "Any"


def _get_sequence_typehint(sequence: Sequence[Any]) -> str:
    # Generate a type hint for a sequence collection.
    return _get_multitype_typehint(sequence, ", ") or "Any, ..."


def _get_mixedtype_typehint(collection: Iterable[Any]) -> str:
    # Generate a type hint for a collection with mixed types.
    origin_type = type(collection).__name__
    type_hinter = (
        _get_sequence_typehint
        if isinstance(collection, tuple)
        else _get_iterable_typehint
    )
    arg_type_hint = type_hinter([*collection])
    return f"{origin_type}[{arg_type_hint}]" if arg_type_hint else origin_type


def _get_dict_typehint(dictionary: dict[Any, Any]) -> str:
    # Generate a type hint for a dictionary.
    keys = [*dictionary.keys()]
    values = [*dictionary.values()]
    keys_type_hint = _get_iterable_typehint(keys)
    vals_type_hint = _get_iterable_typehint(values)
    suffix = f"[{keys_type_hint}, {vals_type_hint}]"
    prefix = type(dictionary).__name__
    return f"{prefix}{suffix}"


def _get_ndarray_typehint(array: ndarray[tuple[int, ...], dtype[Any]]) -> str:
    # Generate a type hint for a NumPy ndarray.
    shape = "".join(f"Literal[{n}], " for n in array.shape[:-1])
    shape += f"Literal[{array.shape[-1]}]"
    suffix = f"[tuple[{shape}], dtype[{str(array.dtype)}]]"
    prefix = type(array).__name__  # str(type(array))[8:-2]
    return f"{prefix}{suffix}"


def _get_collection_typehint(collection: Iterable[Any]) -> str:
    # Generate a type hint for a generic collection.
    if isinstance(collection, dict):
        return _get_dict_typehint(collection)
    if isinstance(collection, ndarray):
        return _get_ndarray_typehint(collection)
    return _get_mixedtype_typehint(collection)


def get_typehint(value: Any) -> str:
    """
    Generate a type hint for a given value.

    Parameters
    ----------
    value : Any
        The value to generate a type hint for.

    Returns
    -------
    str
        The type hint for the given value.
    """
    # Generate a type hint for a given value.
    if isinstance(value, Iterable) and not isinstance(value, str):
        return _get_collection_typehint(value)
    return type(value).__name__


def get_annotated(annotation: type) -> str:
    """
    Generate a type hint for a given annotation.

    Converts a type annotation to a simplified string representation by
    removing specific substrings.

    Parameters
    ----------
    annotation : type
        The type annotation to generate a hint for.

    Returns
    -------
    str
        The type hint for the given annotation.
    """
    remove = ["numpy.", "<class '", "'>"]
    hint = str(annotation)
    for tag in remove:
        hint = hint.replace(tag, "")
    return hint



"""
Provide GridSat or GOES-R series dataset product locator.

Classes:
    - ProductLocatorGG: Abstract a GridSat or GOES-R series dataset
      product locator.
"""

from abc import abstractmethod
from collections.abc import Iterable
from datetime import datetime, timezone
from re import Match, findall, fullmatch

from .locator import ProductLocator


class ProductLocatorGG(ProductLocator):
    """
    Abstract a GridSat or GOES-R series dataset product locator.

    This class implements the `ProductLocator` interface and serves as
    a base class for product locators of the GridSat and GOES-R series
    datasets. It also defines common abstract methods that subclasses
    must implement to provide dataset-specific product utilities.

    Instances of this class are responsible for verifying if a given
    filename matches the product filename pattern based on the dataset's
    naming conventions and product specifications by means of its method
    `match()`, and for extracting the corresponding `datetime`
    information from the product's filename through the use of the
    `get_datetime()` method.

    Notes
    -----
    Subclasses must implement the following methods: `get_base_url()`
    and `get_paths()` declared by the `ProductLocator` interface.
    Implementations for abstract methods declared by this class:
    `get_date_format()`, `get_prefix()`, `get_suffix()`, and
    `get_timestamp_pattern()` should also be provided. Refer to their
    individual documentation for details.

    Methods
    -------
    get_date_format() -> str:
        Return the date format specification for the product's filename.
    get_datetime(filename: str) -> datetime:
        Extract the `datetime` from the product's filename.
    get_filename_pattern() -> str:
        Return a regular expression pattern for the product's filename.
    get_prefix():
        Return the prefix for the product's filename.
    get_suffix():
        Return the suffix for the product's filename.
    get_timestamp_pattern() -> str:
        Return the timestamp regex pattern for the product's filename.
    match(filename: str) -> bool:
        Verify if a provided filename matches the required format.
    timestamp_to_datetime(timestamp: str) -> datetime:
        Convert a timestamp string to a UTC `datetime` object.

    Caution
    -------
    Members of this class not defined by the `ProductLocator` interface
    are helper methods and can be considered as implementation details,
    even though they are defined as part of the public API. In future
    releases, these methods may be moved to a private scope, suffer name
    changes, or be removed altogether.
    """

    @abstractmethod
    def get_date_format(self) -> str:
        """
        Return the date format specification for the product's filename.

        Generates and returns the date format specification for the
        product's filename based on the dataset product filename's date
        and time format conventions. The date format specification
        string is used to parse the product's filename and extract the
        `datetime` information.

        Returns
        -------
        str
            The date format specification for the product's filename.
        """

    def get_datetime(self, filename: str) -> datetime:
        """
        Extract the `datetime` from the product's filename.

        This method parses the given filename and convert it to the
        corresponding `datetime` object from the product's filename
        using the dataset's date and time format conventions.

        Parameters
        ----------
        filename : str
            The filename from which to extract the `datetime`.

        Returns
        -------
        datetime
            The `datetime` extracted from the filename.

        Raises
        ------
        ValueError
            If the filename does not match the expected pattern.
        """
        pattern: str = self.get_filename_pattern()
        matches: list[str] = findall(pattern, filename)

        if len(matches) != 1:
            raise ValueError(
                f"Expected 1 timestamp field, found {len(matches)} fields"
            )

        return self.timestamp_to_datetime(matches[0])

    def get_filename_pattern(self) -> str:
        """
        Return a regular expression pattern for the product's filename.

        Generates and returns a regular expression pattern for the
        product's filename based on the prefix, date pattern, and
        suffix created with dataset-specific product information.

        Returns
        -------
        str
            The regular expression pattern for the product's filename
        """
        prefix: str = self.get_prefix()
        suffix: str = self.get_suffix()
        timestamp_pattern: str = self.get_timestamp_pattern()

        return f"{prefix}{timestamp_pattern}{suffix}"

    @abstractmethod
    def get_prefix(self) -> str:
        """
        Return the prefix for the product's filename.

        Generates and returns the prefix for the product's filename
        based on product-specific information like dataset and product's
        name, instrument and origin's identifier, etc.

        Returns
        -------
        str
            The prefix for the product's filename.
        """

    @abstractmethod
    def get_suffix(self) -> str:
        """
        Return the suffix for the product's filename.

        Generates and returns the suffix for the product's filename
        based on product-specific information like product's version
        origin's identifier, and file suffix (extension).

        Returns
        -------
        str
            The suffix for the product's filename.
        """

    @abstractmethod
    def get_timestamp_pattern(self) -> str:
        """
        Return the timestamp regex pattern for the product's filename.

        Generates and returns the timestamp regular expression
        pattern for the product's filename based on the dataset
        product filename's date and time format conventions. The
        timestamp regex pattern is used to extract the substring
        containing the timestamp from the product's filename
        before extracting the `datetime` information.

        Returns
        -------
        str
            The timestamp regex pattern for the product's filename.
        """

    def match(self, filename: str) -> bool:
        """
        Verify if a provided filename matches the required format.

        Checks if the provided filename matches the product's filename
        pattern for the dataset product.

        Parameters
        ----------
        filename : str
            The filename to match against the pattern.

        Returns
        -------
        bool
            True if the filename matches the pattern, False otherwise.
        """
        pattern: str = self.get_filename_pattern()
        match: Match[str] | None = fullmatch(pattern, filename)

        return match is not None

    def timestamp_to_datetime(self, timestamp: str) -> datetime:
        """
        Convert a timestamp string to a UTC `datetime` object.

        Parses and converts the provided timestamp string to a
        `datetime` object.

        Notes
        -----
        The timestamp string extracted from the product's filename are
        (or assumed to be) always in UTC timezone. Consumer of product
        utilities should be aware of this assumption. The user could
        convert the `datetime` object to the desired timezone if needed.

        The framework raises `ValueError` if the timestamp does not
        match the expected format or if the format specification is
        ill-formed.

        Parameters
        ----------
        timestamp : str
            The timestamp string to convert to a `datetime` object.

        Returns
        -------
        datetime
            The converted `datetime` object in UTC timezone.
        """
        # File dates are always in UTC.
        file_timestamp: str = f"{timestamp}+0000"
        date_format: str = self.get_date_format()
        file_date_format: str = f"{date_format}%z"
        file_date: datetime = datetime.strptime(
            file_timestamp, file_date_format
        )

        return file_date.astimezone(timezone.utc)

    @staticmethod
    def _validate_entity(
        name: str, entity: str, available_entities: Iterable[str]
    ) -> None:
        if entity not in available_entities:
            supported_entities: list[str] = sorted(available_entities)
            supported_ids = "', '".join(supported_entities)
            raise ValueError(
                f"Invalid {name} ID: '{entity}'. "
                f"Available {name} IDs: '{supported_ids}'"
            )

    @staticmethod
    def _validate_set(
        name: str,
        entities: str | Iterable[str],
        available_entities: Iterable[str],
    ) -> None:
        if not set(entities).issubset(available_entities):
            invalid_entities = set(entities) - set(available_entities)
            supported_entities: list[str] = sorted(available_entities)
            invalid_ids = "', '".join(invalid_entities)
            supported_ids = "', '".join(supported_entities)
            raise ValueError(
                f"Invalid {name} IDs: '{invalid_ids}'. "
                f"Available {name} IDs: '{supported_ids}'"
            )

    @classmethod
    def _validate_channels(
        cls: type["ProductLocatorGG"],
        channel: str | Iterable[str],
        available_channels: Iterable[str],
    ) -> None:
        cls._validate_set("channel", channel, available_channels)

    @classmethod
    def _validate_datasource(
        cls: type["ProductLocatorGG"],
        datasource: str,
        available_datasources: Iterable[str],
    ) -> None:
        cls._validate_entity("datasource", datasource, available_datasources)

    @classmethod
    def _validate_instrument(
        cls: type["ProductLocatorGG"],
        instrument: str,
        available_instruments: Iterable[str],
    ) -> None:
        cls._validate_entity("instrument", instrument, available_instruments)

    @classmethod
    def _validate_level(
        cls: type["ProductLocatorGG"], level: str, available_levels: set[str]
    ) -> None:
        cls._validate_entity("level", level, available_levels)

    @classmethod
    def _validate_origin(
        cls: type["ProductLocatorGG"],
        origin: str,
        available_origins: Iterable[str],
    ) -> None:
        cls._validate_entity("origin", origin, available_origins)

    @classmethod
    def _validate_product(
        cls: type["ProductLocatorGG"],
        name: str,
        available_products: Iterable[str],
    ) -> None:
        cls._validate_entity("product", name, available_products)

    @classmethod
    def _validate_scene(
        cls: type["ProductLocatorGG"],
        scene: str,
        available_scenes: Iterable[str],
    ) -> None:
        cls._validate_entity("scene", scene, available_scenes)
